/* --------------------------------------------------------------
    Signed integers with unlimited range (version 1.2b).

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2009 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifndef _CBIGNUM_H
#define _CBIGNUM_H

#ifndef __cplusplus
#error  Must use C++ for the type cBigNumber.
#endif//__cplusplus

#ifndef CTTY_H
#include "Ctty.h"
#endif//CTTY_H

#ifndef EXARRAY_H
#include "Exarray.h"
#include "Exdebug.h"
#endif//EXARRAY_H

#if 0
#define _CBIGNUM_MT
#endif

//================================================
//      Internal representation of cBigNumber.
//================================================
//      Unlimited number is stored in array of type long.
//      Word 0 contains number of long words, required for
//      storing of complementary code of number, the next
//      words contain complementary code in order from low
//      words to high words. Sign resides in the last word.
//      Method length() tells number of words in the code.
//
//      Minimal number of long word are used for storing,
//      but no less then one word:
//      - if length()>1, high word differs from sign extension
//        of word preceding the high;
//      - zero is represented by single 0-word.
//      We'll call this representation to be normalized.
//
//      Normalization is not mandatory. Methods allow for any
//      number of words in the complementary code, starting from 0.
//      Obtaining of normalized result is guaranteed if all
//      operands are normalized.
//
//      From managing of long array we use class exblock<long>,
//      which object contains reference to array with code of number
//      and also field with overall number of long words, available
//      for recording if number. Number of available words may be
//      be greater then number of occupied words as this is necessary
//      for arithmetical operation, which may require for increasing
//      of size of number.
//
//      Resizing of array is carried out by means of the following
//      methods:
//      checkexpand(n) - increase number of available words and
//                       reserve some memory for further expansion.
//      checkindex(n)  - decrease number of available words but
//                       do not free reserved memory.
//      Here number of words does not include word with index 0.
//      Note that array may reserve more words than it is required.
//
//      If macro NCHECKPTR is defined, method checkexpand() increases
//      number of available words uo to number of actually reserved
//      words, whereas method checkindex() does nothing.
//
//      If macro NCHECKPTR is not defined, then methods checkexpand()
//      and checkindex() sets number of available words to be exactly
//      same as indicated.
//
//      Hierarchy of classes:
//
//      exblock<long>           // Array with variable length.
//      |
//      +- cBigNumber           // Unlimited number.
//         |
//         +- cBigTemp          // Unlimited number in the stack.
//            |
//            +- cBigInc        // Postfix increment.
//            +- cBigDec        // Postfix decrement.
//            +- cBigPos        // Unary +.
//            +- cBigNeg        // Unary -.
//            +- cBigAbs        // Absolute value.
//            +- cBigUnsign     // Unsigned value.
//            +- cBigCompl      // Bit-wise inversion.
//            +- cBigXor        // Bit-wise sum by module 2.
//            +- cBigAnd        // Bit-wise conjunction.
//            +- cBigOr         // Bit-wise disjunction.
//            +- cBigAdd        // Addition.
//            +- cBigSub        // Subtraction.
//            +- cBigMul        // Multiplication.
//            +- cBigDiv        // Division.
//            +- cBigDivMod     // Division with module.
//            +- cBigMod        // Module.
//            +- cBigShl        // Left shift.
//            +- cBigShr        // Right shift.
//            +- cBigPow        // Power.
//            +- cBigPowMod     // Power by module.
//            +- cBigSqrt       // Square root.
//            +- cBigSqrtRm     // Square root with remainder.
//            +- cBigBits       // Number of meaning bits.
//            +- cBigExBits     // Number of meaning low 0-bits.
//            +- cBigRandom     // Random value.

//================================================
//      Typedef for string conversion.
//================================================

typedef exarray<char> cBigString;

//================================================
//      Main class cBigNumber.
//================================================

class cBigNumber: public exblock<long>
{
protected:                      // Optimized method for creating of number
                                // with minimal size and given initial value.
  void  create (long b = 0)     // Called from some constructors instead of
  {                             // checkexpand (1).
    e = (long *) exmalloc (EXCALCBLOCKSIZE_1);
    e [0] = 1; e [1] = b;
    len = max_size (EXCALCBLOCKSIZE_1);
    checkindex (1);
  }
                                // Constructor for cBigTemp.
  cBigNumber (exblockstack<long> &stack) { stack.get (*this); }

public:

//      Constructors.                                   // Initial value:
  cBigNumber ()                 { create (0);   }       // Zero.
  cBigNumber (long b)           { create (b);   }       // Number.
  cBigNumber (const cBigNumber&);                       // Copying.
  cBigNumber (const char* psz, unsigned radix);         // String.

//      Destructor.
  ~cBigNumber()                 { exfree (e);   }

//      Internal representation of number.
  size_t      length  () const  { return (size_t)(*e);} // Number of words.
  const long* code    () const  { return e + 1;       } // Array of words.
  const long& loword  () const  { return e [1];       } // Low word.
  const long& hiword  () const  { return e [length()];} // High word.
  size_t      isfit   () const;                         // Normalized?
  size_t      words   () const;                         // Count meaning words.
  size_t      exwords () const;                         // Count low 0-words.

//      Information on number.
  long        bits    () const;                         // Count meaning bits.
  cBigNumber& setbits   (const cBigNumber&);
  cBigNumber& setbits   (long);
  long        exbits  () const;                         // Count low 0-bits.
  cBigNumber& setexbits (const cBigNumber&);
  cBigNumber& setexbits (long);

//      Service conversions.
//      Number does not change but its internal representation may change.
  cBigNumber& fit   ();                                 // Normalize.
  cBigNumber& tab   ();                                 // Table of shifts.
  cBigNumber& smp   ();                                 // Prepare mulsmp().
  cBigNumber& gc    ();                                 // Optimize memory.
  cBigNumber& pack  ();                                 // Truncate memory.
const cBigNumber& fit   () const { return ((cBigNumber*)this)->fit();   }
const cBigNumber& tab   () const { return ((cBigNumber*)this)->tab();   }
const cBigNumber& smp   () const { return ((cBigNumber*)this)->smp();   }
const cBigNumber& gc    () const { return ((cBigNumber*)this)->gc();    }
const cBigNumber& pack  () const { return ((cBigNumber*)this)->pack();  }

//      Comparison operations. Returns -1 = less, 0 = equal, 1 = greater.
  int         comp  (const cBigNumber&) const;          // Compare to
  int         comp  (long)              const;          // given number.
  int         comp0 ()                  const;          // Compare to 0.

//      Unary operations with accumulation.
  cBigNumber& neg   ()  { return setneg  (*this); }     // Inversion of sign.
  cBigNumber& abs   ();                                 // Absolute value.
  cBigNumber& unsign();                                 // Unsigned value.
//cBigNumber& compl ()  { return setcompl(*this); }     // Bit-wise inversion.
  cBigNumber& operator ++();                            // Prefix increment.
  cBigNumber& operator --();                            // Prefix decrement.
  cBigNumber& mul2  ();                                 // Multiplication to 2.
  cBigNumber& div2  ();                                 // Division to 2.
  cBigNumber& pow2  ();                                 // Square.
  cBigNumber& sqrt  ();                                 // Square root.

//      Assign either copy of number or result of operation.
//      Operands are passed as parameters.
  cBigNumber& set         (const cBigNumber&);          // Copying.
  cBigNumber& set         (long);
  cBigNumber& setneg      (const cBigNumber&);          // Inversion of sign.
  cBigNumber& setneg      (long);
  cBigNumber& setabs      (const cBigNumber&);          // Absolute value.
  cBigNumber& setabs      (long);
  cBigNumber& setunsign   (const cBigNumber&);          // Unsigned value.
  cBigNumber& setunsign   (long);
  cBigNumber& setcompl    (const cBigNumber&);          // Bit-wise inversion.
  cBigNumber& setcompl    (long);
  cBigNumber& setxor (const cBigNumber&, const cBigNumber&);    // Bit-wise
  cBigNumber& setxor (const cBigNumber&, long);                 // sum by
  cBigNumber& setxor (long, const cBigNumber&);                 // module 2.
  cBigNumber& setxor (long, long);
  cBigNumber& setand (const cBigNumber&, const cBigNumber&);    // Bit-wise
  cBigNumber& setand (const cBigNumber&, long);                 // conjunction.
  cBigNumber& setand (long, const cBigNumber&);
  cBigNumber& setand (long, long);
  cBigNumber& setor  (const cBigNumber&, const cBigNumber&);    // Bit-wise
  cBigNumber& setor  (const cBigNumber&, long);                 // disjunction.
  cBigNumber& setor  (long, const cBigNumber&);
  cBigNumber& setor  (long, long);
  cBigNumber& setadd (const cBigNumber&, const cBigNumber&);    // Addition.
  cBigNumber& setadd (const cBigNumber&, long);
  cBigNumber& setadd (long, const cBigNumber&);
  cBigNumber& setadd (long, long);
  cBigNumber& setsub (const cBigNumber&, const cBigNumber&);    // Subtraction.
  cBigNumber& setsub (const cBigNumber&, long);
  cBigNumber& setsub (long, const cBigNumber&);
  cBigNumber& setsub (long, long);
  cBigNumber& setmul (const cBigNumber&, const cBigNumber&);    // Multipli-
  cBigNumber& setmul (const cBigNumber&, long);                 // cation.
  cBigNumber& setmul (long, const cBigNumber&);
  cBigNumber& setmul (long, long);
  cBigNumber& setdiv (const cBigNumber&, const cBigNumber&);    // Division.
  cBigNumber& setdiv (const cBigNumber&, long);
  cBigNumber& setdiv (long, const cBigNumber&);
  cBigNumber& setdiv (long, long);
  cBigNumber& setmod (const cBigNumber&, const cBigNumber&);    // Module.
  cBigNumber& setmod (const cBigNumber&, long);
  cBigNumber& setmod (long, const cBigNumber&);
  cBigNumber& setmod (long, long);
  cBigNumber& setshl (const cBigNumber&, const cBigNumber&);    // Left
  cBigNumber& setshl (const cBigNumber&, long);                 // shift.
  cBigNumber& setshl (long, const cBigNumber&);
  cBigNumber& setshl (long, long);
  cBigNumber& setshr (const cBigNumber&, const cBigNumber&);    // Right
  cBigNumber& setshr (const cBigNumber&, long);                 // shift.
  cBigNumber& setshr (long, const cBigNumber&);
  cBigNumber& setshr (long, long);
  cBigNumber& setpow (const cBigNumber&, const cBigNumber&);    // Power.
  cBigNumber& setpow (const cBigNumber&, long);
  cBigNumber& setpow (long, const cBigNumber&);
  cBigNumber& setpow (long, long);
  cBigNumber& setpowmod (const cBigNumber&, const cBigNumber&,  // Power by
                         const cBigNumber&);                    // module.
  cBigNumber& setpowmod (const cBigNumber&, const cBigNumber&, long);
  cBigNumber& setpowmod (const cBigNumber&, long, const cBigNumber&);
  cBigNumber& setpowmod (const cBigNumber&, long, long);
  cBigNumber& setpowmod (long, const cBigNumber&, const cBigNumber&);
  cBigNumber& setpowmod (long, const cBigNumber&, long);
  cBigNumber& setpowmod (long, long, const cBigNumber&);
  cBigNumber& setpowmod (long, long, long);
  cBigNumber& setsqrt   (const cBigNumber&);                    // Square root.
  cBigNumber& setsqrt   (long);
  cBigNumber& setrandom (unsigned long (*p)(), unsigned long);  // Random.

//      Assignment.
  cBigNumber& operator =  (const cBigNumber& b) { return set (b);       }
  cBigNumber& operator =  (long b)              { return set (b);       }

//      Binary operations with accumulation.
//      Second operand are passed as parameter.
  cBigNumber& operator ^= (const cBigNumber&);  // Bit-wise sum by mod 2
  cBigNumber& operator ^= (long);
  cBigNumber& operator &= (const cBigNumber&);  // Bit-wise conjunction.
  cBigNumber& operator &= (long);
  cBigNumber& operator |= (const cBigNumber&);  // Bit-wise disjunction.
  cBigNumber& operator |= (long);
  cBigNumber& operator += (const cBigNumber&);  // Addition.
  cBigNumber& operator += (long);
  cBigNumber& operator -= (const cBigNumber&);  // Subtraction.
  cBigNumber& operator -= (long);
  cBigNumber& operator *= (const cBigNumber&);  // Multiplication.
  cBigNumber& operator *= (long);
  cBigNumber& operator /= (const cBigNumber&);  // Division.
  cBigNumber& operator /= (long);
  cBigNumber& operator %= (const cBigNumber&);  // Module.
  cBigNumber& operator %= (long);
  cBigNumber& operator<<= (const cBigNumber&);  // Left shift.
  cBigNumber& operator<<= (long);
  cBigNumber& operator>>= (const cBigNumber&);  // Right shift.
  cBigNumber& operator>>= (long);
  cBigNumber& pow         (const cBigNumber&);  // Power.
  cBigNumber& pow         (long);

//      Ternary operations with accumulation.
//      Second and third operands are passed as parameters.
  cBigNumber& powmod      (const cBigNumber&, const cBigNumber&);
  cBigNumber& powmod      (const cBigNumber&, long);
  cBigNumber& powmod      (long, const cBigNumber&);
  cBigNumber& powmod      (long, long);

//      Copying combined with long-words left or right shift.
  cBigNumber& set         (const cBigNumber&, size_t k);
  cBigNumber& set         (long, size_t k);
  cBigNumber& setr        (const cBigNumber&, size_t k);

//      Accumulation combined with long-words left shift.
  cBigNumber& add         (const cBigNumber&, size_t k = 0);
  cBigNumber& add         (long, size_t k = 0);
  cBigNumber& sub         (const cBigNumber&, size_t k = 0);
  cBigNumber& sub         (long, size_t k = 0);

//      Combined operation of multiplication with accumulation.
//      Operands are passed as parameters, result of multiplication
//      is accumulated.
  cBigNumber& addmul      (const cBigNumber&, const cBigNumber&);
  cBigNumber& addmul      (const cBigNumber&, long);
  cBigNumber& addmul      (long, const cBigNumber&);
  cBigNumber& addmul      (long, long);
  cBigNumber& submul      (const cBigNumber&, const cBigNumber&);
  cBigNumber& submul      (const cBigNumber&, long);
  cBigNumber& submul      (long, const cBigNumber&);
  cBigNumber& submul      (long, long);

//      Combined operation of division with module.
//      Module is assigned to dividend, which must not overlap with result.
  cBigNumber& setdivmod   (cBigNumber&, const cBigNumber&);
  cBigNumber& setdivmod   (cBigNumber&, long);

//      Combined operation of square root with remainder.
//      Remainder is assigned to operand, which must not overlap with result.
  cBigNumber& setsqrtrm   (cBigNumber&);

//      Multiplication with accumulation of preliminary prepared operands.
//      Multiplicand must contain table of shifts, prepared by tab().
//      Multiplier must be non-negative.
//      Operands must not overlap with buffer of result.
  cBigNumber& addmultab   (const cBigNumber&, const cBigNumber&, size_t k = 0);
  cBigNumber& addmultab   (const cBigNumber&, long, size_t k = 0);
  cBigNumber& submultab   (const cBigNumber&, const cBigNumber&, size_t k = 0);
  cBigNumber& submultab   (const cBigNumber&, long, size_t k = 0);
//      Multiplicand must be prepared by either smp() or tab().
//      Multiplier must be non-negative.
//      Operands must not overlap with buffer of result.
  cBigNumber& addmulsmp   (const cBigNumber&, const cBigNumber&, size_t k = 0);
  cBigNumber& addmulsmp   (const cBigNumber&, long, size_t k = 0);
  cBigNumber& submulsmp   (const cBigNumber&, const cBigNumber&, size_t k = 0);
  cBigNumber& submulsmp   (const cBigNumber&, long, size_t k = 0);

//      Division of preliminary prepared operands.
//      Operands must have identical signs. Dividend must be normalized,
//      divider must contain table of shifts, prepared by tab().
//      and must not overlap with buffer of result.
  cBigNumber& setdivtab (const cBigNumber&, const cBigNumber&); // Division.
  cBigNumber& setdivtab (long, const cBigNumber&);
  cBigNumber& setmodtab (const cBigNumber&, const cBigNumber&); // Module.
  cBigNumber& setmodtab (long, const cBigNumber&);

//      Division with module of preliminary prepared operands.
//      Operands must have identical signs, must not overlap with buffer of
//      result and each other. Dividend must be normalized,
//      divider must contain table of shifts, prepared by tab().
  cBigNumber& setdivmodtab    (cBigNumber&, const cBigNumber&); // Division
                                                                // with module.
//      Messages.
static  const char* pszConverting;
static  const char* pszFormatting;

//      String conversions.
#define MAX_RADIX 16
static  const char  numtochar [2] [MAX_RADIX + 1];      // Undocumented
static        char  chartonum [UCHAR_MAX + 1];          // internal arrays.
static  const long          maxradix[][4];
static  exarray<cBigNumber> powradix [MAX_RADIX + 1];
static  size_t  SqrTab  (unsigned radix, size_t len = 0, size_t max_np = 0);
  cBigNumber& setunsign (const EXPTR(char) ps,          // Conversion of array
                         unsigned radix,                // to unsigned number.
                         size_t   width);
  cBigNumber& set       (const char* psz,               // Conversion of string
                         unsigned radix = 10);          // to number.
#define cBigNumber_uppercase    0x00040000L
#define cBigNumber_showbase     0x00080000L
#define cBigNumber_showpos      0x00200000L
#define cBigNumber_unsign       0x80000000L
  char*       toa       (cBigString& buf_,
                         unsigned radix = 10,           // Preparing string in
                         size_t   width = 0,            // extendable buffer.
                         long     fill  = 0)    const;
#ifndef _CBIGNUM_MT
#define _CBIGNUM_TOATMP
#endif//_CBIGNUM_MT
#ifdef  _CBIGNUM_TOATMP
  char*       toatmp    (unsigned radix = 10,           // Preparing string in
                         size_t   width = 0,            // the static buffer
                         long     fill  = 0)    const;  // (deprecated).
#endif//_CBIGNUM_TOATMP
  void        dump      ()  const;                      // Dump array of words.
  void        info      ()  const;                      // Short info of words.

//      Conversion to integer with check of range.
  void        erange    ()  const;                      // Range error.
  long        tolong    ()  const;
  int         toint     ()  const;
  short       toshort   ()  const;

//      Conversions for logical C operations.
  operator const void * ()  const       { return (comp0()? e: 0);       }
  int        operator ! ()  const       { return (comp0() == 0);        }

//      Control of division.
static  void  maskdiv0  (int);          // Mask division by 0.
static  int   testdiv0  ();             // Is where division by 0?

//      Access to internal static buffers (deprecated).
#ifndef _CBIGNUM_MT
static  cBigNumber& lastdivmod ();      // Module of last division.
static  cBigNumber& lastrootrm ();      // Remainder of last square root.
#endif//_CBIGNUM_MT
};

//================================================
//      Basic test for primality.
//================================================

//      Strong Probable Primality test for base b.
int     b_SPRP      (const cBigNumber& n, const cBigNumber& b);

//================================================
//      Integral tests for primality (prime.cpp).
//================================================

//      Strong Probable Primality test.
int     SPRP        (const cBigNumber&);
int     SPRP        (const cBigNumber&, cBigNumber&);
int FastSPRP        (const cBigNumber&);
int FastSPRP        (const cBigNumber&, cBigNumber&);
int LastSPRP        (const cBigNumber&);
int LastSPRP        (const cBigNumber&, cBigNumber&);

//      Factor Primality test.
int TestFactor      (const cBigNumber&);
int TestFactor      (const cBigNumber&, cBigNumber&);
int TestSmallFactor (const cBigNumber&);
int TestSmallFactor (const cBigNumber&, cBigNumber&);
int TestLargeFactor (const cBigNumber&);
int TestLargeFactor (const cBigNumber&, cBigNumber&);

//      Combined (Factor + Strong) Primality test.
int IsPrime         (const cBigNumber&);
int IsProvedPrime   (const cBigNumber&);
int IsMillerPrime   (const cBigNumber&);
int IsStrongPrime   (const cBigNumber&);

//      Old function, compatible with version 1.1a
#ifdef  _CBIGNUM_HASFACTOR
inline int HasFactor(const cBigNumber& n) { return TestFactor(n);   }
#endif//_CBIGNUM_HASFACTOR

//================================================
//      Compiler independent stream output.
//================================================

cTTY& operator << (cTTY& os, const cBigNumber& b);
void operator <<= (cTTY& os, const cBigNumber& b);

#ifdef  __cplusplus
extern "C" {
#endif//__cplusplus

extern cTTY cBigNumberMessages;         // Handler for messages.
extern cTTY cBigNumberProgress;         // Handler for progress indicator.

#ifdef  __cplusplus
}
#endif//__cplusplus

//================================================
//      Class to convert number to hex string.
//================================================

class cHexDump
{
  char buffer [CHAR_BIT * sizeof (long) * 3 / 8];
public:
  cHexDump (unsigned long n);
  operator const char*() { return (buffer); }
        // const avoids troubles with Borland C++
};

//================================================
//      Auxiliary class for passing of long
//      number to basic functions.
//================================================

class _cBigLong
{
  long e [2];           // Work array.

public:                 // Assignment to work array.

  _cBigLong         (long n = 0) { e [0] = 1; e [1] = n;        }

  _cBigLong& operator = (long n) { e [1] = n; return *this;     }

#ifdef  NCHECKPTR

//      Conventional pointer for addressing without check of bounds.

  long*       base()            { return (e);                   }

#else //NCHECKPTR

//      Restricted pointer for addressing with check of bounds:
//      on passing of index outside of array bounds function
//      assigned to ::exalloc_status.range_handler is called.

  operator    exptr<long>()     { return exptr<long> (e, 2);    }

#endif//NCHECKPTR
};

#ifdef  __cplusplus
extern "C" {
#endif//__cplusplus

//================================================
//      Service variables.
//================================================

extern  size_t      cBigNumberMaskDiv0;     // Mask for division by 0.
#ifndef _CBIGNUM_MT
extern  cBigNumber  cBigNumberLastDivMod;   // Module of last division.
extern  cBigNumber  cBigNumberLastRootRm;   // Module of last square root.
#endif//_CBIGNUM_MT

//================================================
//      Basic functions.
//================================================

void    cBigNumberDump    (const EXPTR(long) p1);
void    cBigNumberInfo    (const EXPTR(long) p1);
void    cBigNumberERange  (const EXPTR(long) p1);
size_t  cBigNumberWords   (const EXPTR(long) p1);
int     cLongBits         (long);
int     cULongBits        (unsigned long);
long    cBigNumberBits    (const EXPTR(long) p1);
size_t  cBigNumberExWords (const EXPTR(long) p1);
int     cLongExBits       (long);
int     cLongExactLog2    (long);
long    cBigNumberExBits  (const EXPTR(long) p1);
size_t  cBigNumberIsFit   (const EXPTR(long) p1);
size_t _cBigNumberFit           (EXPTR(long) p1);
size_t _cBigNumberFitTo         (EXPTR(long) p1, size_t n);
int     cBigNumberComp    (const EXPTR(long) p1, const EXPTR(long) p2);
int     cBigNumberCompHigh(const EXPTR(long) p1, const EXPTR(long) p2);
void    cBigNumberCopy    (const EXPTR(long) p1, EXPTR(long) p);
void    cBigNumberCompl   (const EXPTR(long) p1, EXPTR(long) p);
void    cBigNumberCopyShl (const EXPTR(long) p1, size_t k1, EXPTR(long) p);
void    cBigNumberCopyShr (const EXPTR(long) p1, size_t k1, EXPTR(long) p);
void    cBigNumberCutOut  (const EXPTR(long) p1, size_t k1, size_t n,
                                 EXPTR(long) p);
size_t _cBigNumberNeg     (const EXPTR(long) p1, EXPTR(long) p);
size_t _cBigNumberMInc          (EXPTR(long) p1);
size_t _cBigNumberMDec          (EXPTR(long) p1);
size_t _cBigNumberMMul2         (EXPTR(long) p1);
size_t _cBigNumberMDiv2         (EXPTR(long) p1);
size_t _cBigNumberMul2    (const EXPTR(long) p1, EXPTR(long) p);
size_t _cBigNumberDiv2    (const EXPTR(long) p1, EXPTR(long) p);
void    cBigNumberTab           (EXPTR(long) p1);
size_t _cBigNumberXor     (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
size_t _cBigNumberAnd     (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
size_t _cBigNumberOr      (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
size_t _cBigNumberAdd     (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
size_t _cBigNumberSub     (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
size_t _cBigNumberSubS    (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
size_t _cBigNumberMAdd          (EXPTR(long) p1, const EXPTR(long) p2);
size_t _cBigNumberMSub          (EXPTR(long) p1, const EXPTR(long) p2);
size_t  cBigNumberSkipLow0(const EXPTR(long) p1);
void    cBigNumberMAddMulM      (EXPTR(long) p1, const EXPTR(long) p2,
                                 unsigned long l2, size_t k2);
void    cBigNumberMSubMulM      (EXPTR(long) p1, const EXPTR(long) p2,
                                 unsigned long l2, size_t k2);
void    cBigNumberMAddM         (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k2);
void    cBigNumberMSubM         (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k2);
size_t _cBigNumberMSubD         (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k2);
size_t _cBigNumberMMul2M        (EXPTR(long) p1);
size_t _cBigNumberMDiv2D        (EXPTR(long) p1);
void    cBigNumberMShrM         (EXPTR(long) p1);
void    cBigNumberMShlD         (EXPTR(long) p1);
void    cBigNumberMAddShl       (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k2);
void    cBigNumberMSubShl       (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k2);
void    cBigNumberMAddMulShl    (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k, EXPTR(long) p);
void    cBigNumberMSubMulShl    (EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k, EXPTR(long) p);
void    cBigNumberMAddMulShlTab
                          (const EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k, EXPTR(long) p);
void    cBigNumberMSubMulShlTab
                          (const EXPTR(long) p1, const EXPTR(long) p2,
                                 size_t k, EXPTR(long) p);
void    cBigNumberMAddMulShlKar (EXPTR(long) p1, EXPTR(long) p2,
                                 size_t k, EXPTR(long) p);
void    cBigNumberMSubMulShlKar (EXPTR(long) p1, EXPTR(long) p2,
                                 size_t k, EXPTR(long) p);
void    cBigNumberMAddMul (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
void    cBigNumberMSubMul (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
void    cBigNumberMul     (const EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
void    cLongMul                (long, long, long p [2]);
void    cULongMul      (unsigned long, unsigned long, unsigned long p [2]);
void    cBigNumberDiv0();
void    cBigNumberMModDivTab    (EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
void    cBigNumberMModDiv       (EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
void    cBigNumberMModTab       (EXPTR(long) p1, const EXPTR(long) p2);
void    cBigNumberMMod          (EXPTR(long) p1, const EXPTR(long) p2);
void    cBigNumberPow           (EXPTR(long) p1, const EXPTR(long) p2,
                                 EXPTR(long) p);
void    cBigNumberPowMod        (EXPTR(long) p1, const EXPTR(long) p2,
                                                 const EXPTR(long) mod,
                                 EXPTR(long) p);
void    cBigNumberMRmSqrt       (EXPTR(long) p1, EXPTR(long) p);
void    cBigNumberRandom  (unsigned long (*pfnRand)(), unsigned long lBits,
                                 EXPTR(long) p);
#ifdef  __cplusplus
}
#endif//__cplusplus

//================================================
//      Basic inline functions.
//================================================

inline  size_t  cBigNumberNeg     (const EXPTR(long) p1, EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberNeg (p1, p));
}

inline  size_t  cBigNumberMInc          (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberMInc (p1));
}

inline  size_t  cBigNumberMDec          (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberMDec (p1));
}

inline  size_t  cBigNumberMMul2         (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberMMul2 (p1));
}

inline  size_t  cBigNumberMDiv2         (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberMDiv2 (p1));
}

inline  size_t  cBigNumberMul2   (const EXPTR(long) p1, EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberMul2 (p1, p));
}

inline  size_t  cBigNumberDiv2   (const EXPTR(long) p1, EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberDiv2 (p1, p));
}

inline  size_t  cBigNumberFit          (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberFit (p1));
}

inline  size_t  cBigNumberFitTo        (EXPTR(long) p1, size_t n)
{
  return (size_t)(*(p1)=(long)_cBigNumberFitTo (p1, n));
}

inline  size_t  cBigNumberXor    (const EXPTR(long) p1, const EXPTR(long) p2,
                                        EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberXor (p1, p2, p));
}

inline  size_t  cBigNumberAnd    (const EXPTR(long) p1, const EXPTR(long) p2,
                                        EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberAnd (p1, p2, p));
}

inline  size_t  cBigNumberOr     (const EXPTR(long) p1, const EXPTR(long) p2,
                                        EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberOr  (p1, p2, p));
}

inline  size_t  cBigNumberAdd    (const EXPTR(long) p1, const EXPTR(long) p2,
                                        EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberAdd (p1, p2, p));
}

inline  size_t  cBigNumberSub    (const EXPTR(long) p1, const EXPTR(long) p2,
                                        EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberSub (p1, p2, p));
}

inline  size_t  cBigNumberSubS   (const EXPTR(long) p1, const EXPTR(long) p2,
                                        EXPTR(long) p)
{
  return (size_t)(*(p)=(long)_cBigNumberSubS (p1, p2, p));
}

inline  size_t  cBigNumberMAdd         (EXPTR(long) p1, const EXPTR(long) p2)
{
  return (size_t)(*(p1)=(long)_cBigNumberMAdd (p1, p2));
}

inline  size_t  cBigNumberMSub         (EXPTR(long) p1, const EXPTR(long) p2)
{
  return (size_t)(*(p1)=(long)_cBigNumberMSub (p1, p2));
}

inline  size_t  cBigNumberMSubD        (EXPTR(long) p1, const EXPTR(long) p2,
                                        size_t k2)
{
  return (size_t)(*(p1)=(long)_cBigNumberMSubD (p1, p2, k2));
}

inline  size_t  cBigNumberMMul2M        (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberMMul2M (p1));
}

inline  size_t  cBigNumberMDiv2D        (EXPTR(long) p1)
{
  return (size_t)(*(p1)=(long)_cBigNumberMDiv2D (p1));
}

//================================================
//      Implementation of constructors.
//================================================

inline cBigNumber::cBigNumber (const cBigNumber& b)
{
  e = NULL; len = 0;                                    // Initialization by 0.
  checkexpand (b.length());                             // Set size.
  cBigNumberCopy (EXPTRTYPE(b), EXPTRTYPE(*this));      // Copying.
}

inline cBigNumber::cBigNumber (const char* psz, unsigned radix)
{
  create();                                             // Create number.
  set (psz, radix);                                     // Parse constant.
}

//================================================
//      Implementation of information methods.
//================================================

inline void   cBigNumber::dump()    const
{
  cBigNumberDump (EXPTRTYPE(*this));
}

inline void   cBigNumber::info()    const
{
  cBigNumberInfo (EXPTRTYPE(*this));
}

inline void   cBigNumber::erange()  const
{
  cBigNumberERange (EXPTRTYPE(*this));
}

inline size_t cBigNumber::isfit()   const
{
  return cBigNumberIsFit (EXPTRTYPE(*this));
}

inline size_t cBigNumber::words()   const
{
  return cBigNumberWords (EXPTRTYPE(*this));
}

inline size_t cBigNumber::exwords() const
{
  return cBigNumberExWords (EXPTRTYPE(*this));
}

inline long   cBigNumber::bits()    const
{
  return cBigNumberBits (EXPTRTYPE(*this));
}

inline long   cBigNumber::exbits()  const
{
  return cBigNumberExBits (EXPTRTYPE(*this));
}

//================================================
//      Implementation of comparison operations.
//================================================

inline int  cBigNumber::comp (const cBigNumber& b)  const
{
  return cBigNumberComp (EXPTRTYPE(*this), EXPTRTYPE(b));
}

inline int  cBigNumber::comp (long b)               const
{
  return cBigNumberComp (EXPTRTYPE(*this), EXPTRTYPE(_cBigLong(b)));
}

inline int  cBigNumber::comp0 ()                    const
{
  return cBigNumberComp (EXPTRTYPE(*this), EXPTRTYPE(_cBigLong(0)));
}

//================================================
//      Implementation of unary accumulations.
//================================================

inline cBigNumber& cBigNumber::abs()
{
  if (hiword() < 0L) neg();
  return *this;
}

inline cBigNumber& cBigNumber::unsign()
{
  if (hiword() < 0L)
  {
    size_t len = length() + 1;
    checkexpand (len); e [0] = (long)len; e [len] = 0;
  }
  return *this;
}

//================================================
//      Implementation of assign operations.
//================================================

inline cBigNumber& cBigNumber::set (const cBigNumber& b)
{
  checkexpand (b.length());
  cBigNumberCopy (EXPTRTYPE(b), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::set (long b)
{
  checkexpand (1); e [0] = 1; e [1] = b;
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setneg (const cBigNumber& b)
{
  checkexpand (b.length() + 1);
  cBigNumberNeg (EXPTRTYPE(b), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setneg (long b)
{
  checkexpand (2); e [0] = 1 + (b == LONG_MIN); e [1] = -b; e [2] = 0;
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setabs (const cBigNumber& b)
{
  return (b.hiword()>=0? set (b): setneg (b));
}

inline cBigNumber& cBigNumber::setabs (long b)
{
  if (b < 0) b = -b;
  checkexpand (2); e [0] = 1 + (b == LONG_MIN); e [1] = b; e [2] = 0;
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setunsign (const cBigNumber& b)
{
  set (b); return unsign();
}

inline cBigNumber& cBigNumber::setunsign (long b)
{
  checkexpand (2); e [0] = 1 + (b < 0); e [1] = b; e [2] = 0;
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setcompl (const cBigNumber& b)
{
  checkexpand (b.length());
  cBigNumberCompl (EXPTRTYPE(b), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setcompl (long b)
{
  checkexpand (1); e [0] = 1; e [1] = ~b;
  checkindex (length());
  return *this;
}

//================================================
//      Implementation of binary accumulations.
//================================================

inline cBigNumber& cBigNumber::operator ^= (const cBigNumber& b)
{
  return setxor (*this, b);
}

inline cBigNumber& cBigNumber::operator ^= (long b)
{
  return setxor (*this, b);
}

inline cBigNumber& cBigNumber::operator &= (const cBigNumber& b)
{
  return setand (*this, b);
}

inline cBigNumber& cBigNumber::operator &= (long b)
{
  return setand (*this, b);
}

inline cBigNumber& cBigNumber::operator |= (const cBigNumber& b)
{
  return setor (*this, b);
}

inline cBigNumber& cBigNumber::operator |= (long b)
{
  return setor (*this, b);
}

inline cBigNumber& cBigNumber::operator += (const cBigNumber& b)
{
  return add (b);
}

inline cBigNumber& cBigNumber::operator += (long b)
{
  return add (b);
}

inline cBigNumber& cBigNumber::operator -= (const cBigNumber& b)
{
  return sub (b);
}

inline cBigNumber& cBigNumber::operator -= (long b)
{
  return sub (b);
}

inline cBigNumber& cBigNumber::operator *= (const cBigNumber& b)
{
  return setmul (*this, b);
}

inline cBigNumber& cBigNumber::operator *= (long b)
{
  return setmul (*this, b);
}

inline cBigNumber& cBigNumber::operator /= (const cBigNumber& b)
{
#ifdef  _CBIGNUM_MT
  cBigNumber cBigNumberLastDivMod;
#endif//_CBIGNUM_MT
  swap (cBigNumberLastDivMod);
  return setdivmod (cBigNumberLastDivMod, b);
}

inline cBigNumber& cBigNumber::operator /= (long b)
{
#ifdef  _CBIGNUM_MT
  cBigNumber cBigNumberLastDivMod;
#endif//_CBIGNUM_MT
  swap (cBigNumberLastDivMod);
  return setdivmod (cBigNumberLastDivMod, b);
}

inline cBigNumber& cBigNumber::operator %= (const cBigNumber& b)
{
  fit();
  cBigNumberMMod (EXPTRTYPE(*this), EXPTRTYPE(b));
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::operator %= (long b)
{
  fit();
  cBigNumberMMod (EXPTRTYPE(*this), EXPTRTYPE(_cBigLong(b)));
  checkindex (length());
  return *this;
}

//================================================
//      Implementation of copying with shift.
//================================================

inline cBigNumber& cBigNumber::set (const cBigNumber& b, size_t k)
{
  checkexpand (b.length() + k);
  cBigNumberCopyShl (EXPTRTYPE(b), k, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::set (long b, size_t k)
{
  checkexpand (k + 1);
  cBigNumberCopyShl (EXPTRTYPE(_cBigLong(b)), k, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

inline cBigNumber& cBigNumber::setr (const cBigNumber& b, size_t k)
{
  checkexpand (b.length() > k? b.length() - k: 1);
  cBigNumberCopyShr (EXPTRTYPE(b), k, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

//================================================
//      Implementation of check of range.
//================================================

inline  long    cBigNumber::tolong()  const
{
  if (size() == 0)      return (0);
  if (length() > 1)     erange();
  return (hiword());
}

inline  int     cBigNumber::toint()   const
{
  long n = tolong();
  if ((int) n != n)     erange();
  return ((int)n);
}

inline  short   cBigNumber::toshort() const
{
  long n = tolong();
  if ((short) n != n)   erange();
  return ((short)n);
}

//================================================
//      Implementation of control methods.
//================================================

inline void cBigNumber::maskdiv0 (int mask) { cBigNumberMaskDiv0 = mask << 1; }
inline int  cBigNumber::testdiv0()          { return cBigNumberMaskDiv0 == 1; }

#ifndef _CBIGNUM_MT
inline cBigNumber& cBigNumber::lastdivmod() { return cBigNumberLastDivMod; }
inline cBigNumber& cBigNumber::lastrootrm() { return cBigNumberLastRootRm; }
#endif//_CBIGNUM_MT

//================================================
//      Implementation of compare operations.
//================================================

inline  int  operator == (const cBigNumber& a, const cBigNumber& b)
{
  return (a.comp(b) == 0);
}

inline  int  operator == (const cBigNumber& a, long b)
{
  return (a.comp(b) == 0);
}

inline  int  operator == (long b, const cBigNumber& a)
{
  return (a.comp(b) == 0);
}

inline  int  operator != (const cBigNumber& a, const cBigNumber& b)
{
  return (a.comp(b) != 0);
}

inline  int  operator != (const cBigNumber& a, long b)
{
  return (a.comp(b) != 0);
}

inline  int  operator != (long b, const cBigNumber& a)
{
  return (a.comp(b) != 0);
}

inline  int  operator >= (const cBigNumber& a, const cBigNumber& b)
{
  return (a.comp(b) >= 0);
}

inline  int  operator >= (const cBigNumber& a, long b)
{
  return (a.comp(b) >= 0);
}

inline  int  operator >= (long b, const cBigNumber& a)
{
  return (a.comp(b) <= 0);
}

inline  int  operator <= (const cBigNumber& a, const cBigNumber& b)
{
  return (a.comp(b) <= 0);
}

inline  int  operator <= (const cBigNumber& a, long b)
{
  return (a.comp(b) <= 0);
}

inline  int  operator <= (long b, const cBigNumber& a)
{
  return (a.comp(b) >= 0);
}

inline  int  operator >  (const cBigNumber& a, const cBigNumber& b)
{
  return (a.comp(b) > 0);
}

inline  int  operator >  (const cBigNumber& a, long b)
{
  return (a.comp(b) > 0);
}

inline  int  operator >  (long b, const cBigNumber& a)
{
  return (a.comp(b) < 0);
}

inline  int  operator <  (const cBigNumber& a, const cBigNumber& b)
{
  return (a.comp(b) < 0);
}

inline  int  operator <  (const cBigNumber& a, long b)
{
  return (a.comp(b) < 0);
}

inline  int  operator <  (long b, const cBigNumber& a)
{
  return (a.comp(b) > 0);
}

//================================================
//      Inherited classes for implementation
//      of operations which require for creation
//      of temporary objects.
//================================================

class cBigTemp: public cBigNumber               // Number in temporary object.
{
//      Class cBigTemp describes temporary objects
//      for results of arithmetical operations.
//
//      For optimization of memory allocation constructor cBigTemp
//      uses stack of dynamic arrays, which allows to avoid redundant
//      allocations of memory: when object cBigTemp is deleted, pointer
//      to its base array is put to stack; when object cBigTemp is
//      created it receives base array from stack. Memory allocation
//      is required only when the stack if empty.

public:

  static exblockstack<long> stack;              // Stack of dynamic arrays.

#ifndef _CBIGNUM_MT

  cBigTemp (): cBigNumber (stack)   {}          // Constructor.
  ~cBigTemp()  { stack.put (*this);  }          // Destructor.

#else//_CBIGNUM_MT

  cBigTemp ()  {}                               // Constructor.
  ~cBigTemp()  {}                               // Destructor.

#endif//_CBIGNUM_MT

//      Blocking of assignments to objects of class cBigNumber to prevent
//      incorrect assignments to temporary objects, such as (-x) = 10;

private:

  cBigNumber& operator ++ ();                   // Prefix increment.
  cBigNumber& operator -- ();                   // Prefix decrement.
  cBigNumber& operator =  (const cBigNumber&);  // Assignment.
  cBigNumber& operator =  (long);
  cBigNumber& operator ^= (const cBigNumber&);  // Bit-wise sum by mod 2.
  cBigNumber& operator ^= (long);
  cBigNumber& operator &= (const cBigNumber&);  // Bit-wise conjunction.
  cBigNumber& operator &= (long);
  cBigNumber& operator |= (const cBigNumber&);  // Bit-wise disjunction.
  cBigNumber& operator |= (long);
  cBigNumber& operator += (const cBigNumber&);  // Addition.
  cBigNumber& operator += (long);
  cBigNumber& operator -= (const cBigNumber&);  // Subtraction.
  cBigNumber& operator -= (long);
  cBigNumber& operator *= (const cBigNumber&);  // Multiplication.
  cBigNumber& operator *= (long);
  cBigNumber& operator /= (const cBigNumber&);  // Division.
  cBigNumber& operator /= (long);
  cBigNumber& operator %= (const cBigNumber&);  // Module.
  cBigNumber& operator %= (long);
  cBigNumber& operator<<= (const cBigNumber&);  // Left shift.
  cBigNumber& operator<<= (long);
  cBigNumber& operator>>= (const cBigNumber&);  // Right shift.
  cBigNumber& operator>>= (long);
};

struct cBigInc: public cBigTemp                 // Postfix increment.
{
  cBigInc   (cBigNumber& a)                             { set (a); ++a; }
};

struct cBigDec: public cBigTemp                 // Postfix decrement.
{
  cBigDec   (cBigNumber& a)                             { set (a); --a; }
};

struct cBigPos: public cBigTemp                 // Unary +.
{
  cBigPos   (const cBigNumber& a)                       { set (a);      }
  cBigPos   (long a)                                    { set (a);      }
};

struct cBigNeg: public cBigTemp                 // Unary -.
{
  cBigNeg   (const cBigNumber& a)                       { setneg (a);   }
  cBigNeg   (long a)                                    { setneg (a);   }
};

struct cBigAbs: public cBigTemp                 // Absolute value.
{
  cBigAbs   (const cBigNumber& a)                       { setabs (a);   }
  cBigAbs   (long a)                                    { setabs (a);   }
};

struct cBigUnsign: public cBigTemp              // Unsigned value.
{
  cBigUnsign(const cBigNumber& a)                       { setunsign (a);}
  cBigUnsign(long a)                                    { setunsign (a);}
};

struct cBigCompl: public cBigTemp               // Bit-wise inversion.
{
  cBigCompl (const cBigNumber& a)                       { setcompl (a); }
  cBigCompl (long a)                                    { setcompl (a); }
};

struct cBigXor: public cBigTemp                 // Bit-wise sum by module 2.
{
  cBigXor (const cBigNumber& a, const cBigNumber& b)    { setxor (a, b); }
  cBigXor (const cBigNumber& a, long b)                 { setxor (a, b); }
  cBigXor (long a, const cBigNumber& b)                 { setxor (a, b); }
  cBigXor (long a, long b)                              { setxor (a, b); }
};

struct cBigAnd: public cBigTemp                 // Bit-wise conjunction.
{
  cBigAnd (const cBigNumber& a, const cBigNumber& b)    { setand (a, b); }
  cBigAnd (const cBigNumber& a, long b)                 { setand (a, b); }
  cBigAnd (long a, const cBigNumber& b)                 { setand (a, b); }
  cBigAnd (long a, long b)                              { setand (a, b); }
};

struct cBigOr: public cBigTemp                  // Bit-wise disjunction.
{
  cBigOr  (const cBigNumber& a, const cBigNumber& b)    { setor  (a, b); }
  cBigOr  (const cBigNumber& a, long b)                 { setor  (a, b); }
  cBigOr  (long a, const cBigNumber& b)                 { setor  (a, b); }
  cBigOr  (long a, long b)                              { setor  (a, b); }
};

struct cBigAdd: public cBigTemp                 // Addition.
{
  cBigAdd (const cBigNumber& a, const cBigNumber& b)    { setadd (a, b); }
  cBigAdd (const cBigNumber& a, long b)                 { setadd (a, b); }
  cBigAdd (long a, const cBigNumber& b)                 { setadd (a, b); }
  cBigAdd (long a, long b)                              { setadd (a, b); }
};

struct cBigSub: public cBigTemp                 // Subtraction.
{
  cBigSub (const cBigNumber& a, const cBigNumber& b)    { setsub (a, b); }
  cBigSub (const cBigNumber& a, long b)                 { setsub (a, b); }
  cBigSub (long a, const cBigNumber& b)                 { setsub (a, b); }
  cBigSub (long a, long b)                              { setsub (a, b); }
};

struct cBigMul: public cBigTemp                 // Multiplication.
{
  cBigMul (const cBigNumber& a, const cBigNumber& b)    { setmul (a, b); }
  cBigMul (const cBigNumber& a, long b)                 { setmul (a, b); }
  cBigMul (long a, const cBigNumber& b)                 { setmul (a, b); }
  cBigMul (long a, long b)                              { setmul (a, b); }
};

struct cBigDiv: public cBigTemp                 // Division.
{
  cBigDiv (const cBigNumber& a, const cBigNumber& b)    { setdiv (a, b); }
  cBigDiv (const cBigNumber& a, long b)                 { setdiv (a, b); }
  cBigDiv (long a, const cBigNumber& b)                 { setdiv (a, b); }
  cBigDiv (long a, long b)                              { setdiv (a, b); }
};

struct cBigDivMod: public cBigTemp              // Division & module.
{
  cBigDivMod (cBigNumber& a, const cBigNumber& b)       { setdivmod (a, b); }
  cBigDivMod (cBigNumber& a, long b)                    { setdivmod (a, b); }
};

struct cBigMod: public cBigTemp                 // Module.
{
  cBigMod (const cBigNumber& a, const cBigNumber& b)    { setmod (a, b); }
  cBigMod (const cBigNumber& a, long b)                 { setmod (a, b); }
  cBigMod (long a, const cBigNumber& b)                 { setmod (a, b); }
  cBigMod (long a, long b)                              { setmod (a, b); }
};

struct cBigShl: public cBigTemp                 // Left shift.
{
  cBigShl (const cBigNumber& a, const cBigNumber& b)    { setshl (a, b); }
  cBigShl (const cBigNumber& a, long b)                 { setshl (a, b); }
  cBigShl (long a, const cBigNumber& b)                 { setshl (a, b); }
  cBigShl (long a, long b)                              { setshl (a, b); }
};

struct cBigShr: public cBigTemp                 // Right shift.
{
  cBigShr (const cBigNumber& a, const cBigNumber& b)    { setshr (a, b); }
  cBigShr (const cBigNumber& a, long b)                 { setshr (a, b); }
  cBigShr (long a, const cBigNumber& b)                 { setshr (a, b); }
  cBigShr (long a, long b)                              { setshr (a, b); }
};

struct cBigPow: public cBigTemp                 // Power.
{
  cBigPow (const cBigNumber& a, const cBigNumber& b)    { setpow (a, b); }
  cBigPow (const cBigNumber& a, long b)                 { setpow (a, b); }
  cBigPow (long a, const cBigNumber& b)                 { setpow (a, b); }
  cBigPow (long a, long b)                              { setpow (a, b); }
};

struct cBigPowMod: public cBigTemp              // Power by module.
{
  cBigPowMod (const cBigNumber& a, const cBigNumber& b, cBigNumber& mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (const cBigNumber& a, const cBigNumber& b, long mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (const cBigNumber& a, long b, cBigNumber& mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (const cBigNumber& a, long b, long mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (long a, const cBigNumber& b, cBigNumber& mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (long a, const cBigNumber& b, long mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (long a, long b, cBigNumber& mod)
                                                { setpowmod (a, b, mod); }
  cBigPowMod (long a, long b, long mod)
                                                { setpowmod (a, b, mod); }
};

struct cBigSqrt: public cBigTemp                // Square root.
{
  cBigSqrt (const cBigNumber& a)                        { setsqrt (a);   }
  cBigSqrt (long a)                                     { setsqrt (a);   }
};

struct cBigSqrtRm: public cBigTemp              // Square root & remainder.
{
  cBigSqrtRm (cBigNumber& a)                            { setsqrtrm (a); }
};

struct cBigBits: public cBigTemp                // Number of meaning bits.
{
  cBigBits (cBigNumber& a)                              { setbits (a);   }
};

struct cBigExBits: public cBigTemp              // Number of meaning low 0-bits.
{
  cBigExBits (cBigNumber& a)                            { setexbits (a); }
};

struct cBigRandom: public cBigTemp              // Random value.
{
  cBigRandom (unsigned long (*p)(), unsigned long n)    { setrandom (p, n); }
};

//================================================
//      Implementation of unary operations, which
//      require for creation of temporary object.
//================================================

inline cBigInc   operator ++ (cBigNumber &a, int)
{
  return cBigInc (a);
}

inline cBigDec   operator -- (cBigNumber &a, int)
{
  return cBigDec (a);
}

inline cBigPos   operator + (const cBigNumber &a)
{
  return cBigPos (a);
}

inline cBigNeg   operator - (const cBigNumber &a)
{
  return cBigNeg (a);
}

inline cBigCompl operator ~ (const cBigNumber &a)
{
  return cBigCompl (a);
}

//================================================
//      Implementation of binary operations, which
//      require for creation of temporary object.
//================================================

inline cBigXor operator ^ (const cBigNumber& a, const cBigNumber& b)
{
  return cBigXor (a, b);
}

inline cBigXor operator ^ (const cBigNumber& a, long b)
{
  return cBigXor (a, b);
}

inline cBigXor operator ^ (long b, const cBigNumber& a)
{
  return cBigXor (a, b);
}

inline cBigAnd operator & (const cBigNumber& a, const cBigNumber& b)
{
  return cBigAnd (a, b);
}

inline cBigAnd operator & (const cBigNumber& a, long b)
{
  return cBigAnd (a, b);
}

inline cBigAnd operator & (long b, const cBigNumber& a)
{
  return cBigAnd (a, b);
}

inline cBigOr  operator | (const cBigNumber& a, const cBigNumber& b)
{
  return cBigOr  (a, b);
}

inline cBigOr  operator | (const cBigNumber& a, long b)
{
  return cBigOr  (a, b);
}

inline cBigOr  operator | (long b, const cBigNumber& a)
{
  return cBigOr  (a, b);
}

inline cBigAdd operator + (const cBigNumber& a, const cBigNumber& b)
{
  return cBigAdd (a, b);
}

inline cBigAdd operator + (const cBigNumber& a, long b)
{
  return cBigAdd (a, b);
}

inline cBigAdd operator + (long b, const cBigNumber& a)
{
  return cBigAdd (a, b);
}

inline cBigSub operator - (const cBigNumber& a, const cBigNumber& b)
{
  return cBigSub (a, b);
}

inline cBigSub operator - (const cBigNumber& a, long b)
{
  return cBigSub (a, b);
}

inline cBigSub operator - (long a, const cBigNumber& b)
{
  return cBigSub (a, b);
}

inline cBigMul operator * (const cBigNumber& a, const cBigNumber& b)
{
  return cBigMul (a, b);
}

inline cBigMul operator * (const cBigNumber& a, long b)
{
  return cBigMul (a, b);
}

inline cBigMul operator * (long b, const cBigNumber& a)
{
  return cBigMul (a, b);
}

inline cBigDiv operator / (const cBigNumber& a, const cBigNumber& b)
{
  return cBigDiv (a, b);
}

inline cBigDiv operator / (const cBigNumber& a, long b)
{
  return cBigDiv (a, b);
}

inline cBigDiv operator / (long a, const cBigNumber& b)
{
  return cBigDiv (a, b);
}

inline cBigMod operator % (const cBigNumber& a, const cBigNumber& b)
{
  return cBigMod (a, b);
}

inline cBigMod operator % (const cBigNumber& a, long b)
{
  return cBigMod (a, b);
}

inline cBigMod operator % (long a, const cBigNumber& b)
{
  return cBigMod (a, b);
}

inline cBigShl operator << (const cBigNumber& a, const cBigNumber& b)
{
  return cBigShl (a, b);
}

inline cBigShl operator << (const cBigNumber& a, long b)
{
  return cBigShl (a, b);
}

inline cBigShl operator << (long a, const cBigNumber& b)
{
  return cBigShl (a, b);
}

inline cBigShr operator >> (const cBigNumber& a, const cBigNumber& b)
{
  return cBigShr (a, b);
}

inline cBigShr operator >> (const cBigNumber& a, long b)
{
  return cBigShr (a, b);
}

inline cBigShr operator >> (long a, const cBigNumber& b)
{
  return cBigShr (a, b);
}

#endif//_CBIGNUM_H
