/* --------------------------------------------------------------
    Signed integers with unlimited range (version 1.2b).
    Base functions in machine-independent form.

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2009 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifndef _CBIGNUMF_INL
#define _CBIGNUMF_INL

#ifndef _CBIGNUMF_H
#include "Cbignumf.h"
#endif//_CBIGNUMF_H

#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#define DISALLOW_EXARRAY_SHORTTYPE
#include "Cbignum.h"
#define BITS    (CHAR_BIT * sizeof (long))

#ifdef  _MSC_VER
#pragma auto_inline (off)
#endif/*_MSC_VER*/

//================================================
//      Enabling of Cbignumf.cpp optimization.
//================================================
//      _CBIGNUM_OPTIMIZED_COPY     copying.

#ifdef  _CBIGNUM_ASM

#if 1
#define _CBIGNUM_OPTIMIZED_COPY
#endif

#endif//_CBIGNUM_ASM

//================================================
//      Service functions.
//================================================

inline size_t size_max (size_t n1, size_t n2)
{
  return ((n1 > n2)? n1 : n2);
}

inline size_t size_min (size_t n1, size_t n2)
{
  return ((n1 < n2)? n1 : n2);
}

//================================================
//      Use C naming conventions.
//================================================

#ifdef  __cplusplus
extern "C" {
#endif//__cplusplus

//================================================
//      Implementation of cBigNumber operations.
//================================================
//      Unlimited number is stored in array of type long.
//      Word 0 contains number of long words, required for
//      storing of complementary code of number, the next
//      words contain complementary code in order from low
//      words to high words. Sign resides in the last word.
//
//      Operation are implemented as functions of language C.
//      Unlimited numbers are passed through macro EXPTR(long):
//      this is long * in release mode and exptr<long> in
//      debug mode with check of bounds.
//      Input numbers are declared with modifier const.
//
//      For optimization purposes some functions return
//      number of words instead of storing it in word 0.
//      Names of these functions are started from underscore,
//      they are accompanied by inline function which
//      calls function and then fills in word 0.
//
//      Dynamic check if input data is controlled by the
//      following macro:
//      NDEBUG    - turn off check of bounds and assert
//                  checks of input parameters.
//      NCHECKPTR - turn off check of bounds.
//
//      Cycles in functions cBigNumber... are optimized
//      to reduce number of checks and branches when
//      processing one-word numbers. Sequence operations
//      is selected to meet the criteria and optimal
//      loading of Pentium conveyers.

//================================================
//      Diagnostic variables and functions.
//================================================

cTTY    cBigNumberMessages;             // Handler for messages.
cTTY    cBigNumberProgress;             // Handler for progress indicators.

//      Dump of internal representation is performed in order
//      from high words to low words. Each word is preceded by
//      space and outputs in hexadecimal format, in order from
//      high bits to low bits. Output is finished by new line.

void    cBigNumberDump (                        // Dump of code to messages.
                const   EXPTR(long) p1          // Unlimited number.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  for (; n1; n1--) cBigNumberMessages <<" "<< (const char*) cHexDump (p1 [n1]);
  cBigNumberMessages << "\n";
}

//      Short dump of hexadecimal internal code contains
//      the high word followed by total number of bytes.

void    cBigNumberInfo (                        // Information (short dump).
                const   EXPTR(long) p1          // Unlimited number.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  cBigNumberMessages << cHexDump (p1 [n1]);
  if (n1 > 1)
    cBigNumberMessages << "... (" << (long)(n1 * sizeof (long)) << " bytes)\n";
}

void    cBigNumberERange  (                     // Error message.
                const   EXPTR(long) p1          // Unlimited number.
        )
{
  cBigNumberMessages << "cBigNumber value out of range: ";
  cBigNumberInfo (p1);
  abort();
}

//================================================
//      Information functions.
//================================================

//      Meaning words are most senior word, which does not contain
//      only sign and sign extension bits and all younger words.

size_t  cBigNumberWords (                       // Number of meaning words.
                const   EXPTR(long) p1          // Unlimited number.
        )                                       // May be not normalized.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  if (n1)
  {
    long sign = p1 [n1] >> (BITS-1);            // Sign 0 or -1.
    do
    {
      if (p1 [n1] != sign) break;               // Are there meaning bits?
    }
    while (--n1 != 0);
  }
  return (n1);
}

//      In signed number meaning bits are most senior bit, which is
//      distinct from sign bit and all younger bits. For example,
//      0 and -1 contain no meaning bits, 1 and -2 contain 1 meaning bit,
//      127 and -128 contain 7 meaning bits. Absolute value of number
//      containing n meaning bits is not greater than 2 in power n.

int     cLongBits (long num)                    // Number of meaning bits
{                                               // in the range 0..BITS-1.
  int n = BITS;
  num ^= (num >> (BITS-1));                     // Negative number.
  do --n; while ((num <<= 1) > 0);              // Count bits in the word.
  return (n);
}

//      In unsigned number meaning bits are most senior bit which
//      is distinct from 0 and all younger bits.

int     cULongBits (unsigned long num)          // Number of meaning bits
{                                               // in the range 0..BITS.
  int n = BITS;
  while ((long)num > 0) { num <<= 1; --n; }     // Count bits in the word.
  return (n);
}

//      Count meaning bits in unlimited number.
//      If count > LONG_MAX cBigNumberERange() is called.

long    cBigNumberBits (                        // Number of meaning bits.
                const   EXPTR(long) p1          // Unlimited number.
        )                                       // May be not normalized.

{
  size_t n = cBigNumberWords(p1);               // Number of meaning words.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  if (n)
  {                                             // Check for overflow.
    if ((long)n > (long)(LONG_MAX / BITS))
       cBigNumberERange (p1);
    return ((long)(n - 1) * BITS + cULongBits ((unsigned long)
                                    (p1 [n] ^ (p1 [n1] >> (BITS-1)))));
  }
  return (0);
}

//      Low 0-words are all junior meaning zero words.
//      If all words are 0, there are no low 0-words.

size_t  cBigNumberExWords (                     // Number of low 0-words.
                const   EXPTR(long) p1          // Unlimited number.
)                                               // May be not normalized.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  for (size_t n = 0; n < n1; n++)
  {
    if (p1 [n + 1] != 0) return (n);            // Number of 0-words.
  }
  return (0);
}

//      Low 0-bits are all junior meaning zero bits.
//      For example, 0 and odd numbers contain no low 0-bits,
//      2 and -2 contain 1 low 0-bit, 128 and -128 contain
//      7 low 0-bits. Absolute value of number containing
//      n low 0-bits is not less than 2 in power n.

int     cLongExBits (long num)                  // Number of low 0-bits
{                                               // in the range 0..BITS-1.
  int n = 0;
  if (num) while ((num & 1L) == 0) { ++n; num >>= 1; }
  return (n);
}

//      Determine if number is positive and exactly 2 in power n.

int     cLongExactLog2 (long num)               // Either exact log2 or -1.
{                                               // log2 in the range 0..BITS-1.
  int n = cLongExBits (num);
  if ((num >> n) != 1) n = -1;
  return n;
}

//      Count low 0-bits in unlimited number.
//      If count > LONG_MAX cBigNumberERange() is called.

long    cBigNumberExBits (                      // Number of low 0-bits.
                const   EXPTR(long) p1          // Unlimited number.
        )                                       // May be not normalized.

{
  size_t n = cBigNumberExWords(p1);             // Number of low 0-words.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  if (n1)
  {                                             // Check for overflow.
    if ((long)n > (long)((LONG_MAX - (BITS - 1)) / BITS))
      cBigNumberERange (p1);
    return ((long)n * BITS + cLongExBits (p1 [n + 1]));
  }
  return (0);
}

//================================================
//      Normalization and denormalization.
//================================================

//      The number is normalized if minimal number of long words
//      are used for storing, but no less then one word:
//      - if length()>1, high word differs from sign extension
//        of word preceding the high;
//      - zero is represented by single 0-word.

size_t  cBigNumberIsFit (                       // Check of normalization.
                const   EXPTR(long) p1          // Unlimited number.
        )                                       // If number is normalized,
{                                               // return 1 else 0.
  size_t n1 = (size_t)(*p1);                    // Number of words.

  if (n1 > 1) return ((p1 [n1] != (p1 [n1-1] >> (BITS-1))));

  return (n1);                                  // Number of words is 0.
}

//      Normalization to minimal number of words, not less then 1.

size_t  _cBigNumberFit  (                       // Normalization.
                        EXPTR(long) p1          // Buffer with number.
        )                                       // Size of buffer no less 2.
{
  register size_t n1 = (size_t)(*p1);           // Number of words.
  p1 += n1;                                     // Preparing of pointer.
  register long pass;                           // The high word.
  register long lp;                             // The previous word.

  if (n1)                                       // If bits of the high word
  {                                             // and high bit of the word
    lp = *p1;                                   // before high are all 0 or 1
    do continue;                                // then delete the high word.
    while (--n1 != 0 &&
        (pass = lp, lp = p1 [-1], --p1, pass == (lp >> (BITS-1))));
  }                                             // If no words at all
  else { p1 [1] = 0; }                          // then add word 0.

  return (n1 + 1);                              // Number of words.
}
//#define cBigNumberFit(p1)     (size_t)(*(p1)=_cBigNumberFit(p1))

//      Denormalization with increasing of number of words up to
//      given number. If given number is not greater then initial
//      number of words function adds one more word.
//
//      Function returns NON-normalized result.

size_t  _cBigNumberFitTo (                      // Denormalization.
                        EXPTR(long) p1,         // Buffer with number.
                        size_t      n           // Number of words.
        )                                       // Size of buffer no less
{                                               // max (*p1 + 2, n + 1).
  register size_t n1 = (size_t)(*p1);           // Number of words.
  register long pass = p1 [n1] >> (BITS-1);     // Carry.

  do
    p1 [++n1] = pass;                           // Expand the carry.
  while (n1 < n);

  return (n1);                                  // Number of words.
}
//#define cBigNumberFitTo(p1,n) (size_t)(*(p1)=_cBigNumberFitTo(p1,n))

//================================================
//      Functions of comparison.
//================================================

//      Sign bit of more short number expanded to make
//      number of words in both numbers the same.
//      Then the numbers compared word by word in order
//      from high words to low words until mismatch.
//      Comparison of the high words is signed,
//      comparison of lower words is unsigned.

//      For optimization purposes algorithmic equivalent
//      is applied instead of actual expanding of sign bit.

int     cBigNumberComp  (                       // Comparison of p1, p2.
                const   EXPTR(long) p1,         // Number.
                const   EXPTR(long) p2          // Number.
        )                                       // Returns -1 if p1 <  p2
                                                // Returns  0 if p1 == p2
                                                // Returns  1 if p1  > p2
{                                               // p1, p2 may overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words in p1.
  size_t n2 = (size_t)(*p2);                    // Number of words in p2.
  register long lp1 = 0;                        // Current word of p1.
  register long lp2 = 0;                        // Current word of p2.

//      Comparison of sign words.

  if (n1) lp1 = p1 [n1--];                      // Sign word of p1.
  if (n2) lp2 = p2 [n2--];                      // Sign word of p2.

  if (n1 == n2)                                 // Lengths of p1, p2 the same.
  {
    if (lp1 != lp2) goto NotSameSignWord;       // Sign words not the same.
  }
  else if (n1 > n2)                             // If number p1 longer, then
  {                                             // compare it with expansion
    lp2 >>= (BITS-1); ++n2;                     // of sign of p2.
    if (lp1 != lp2) goto NotSameSignWord;       // Sign words not the same.
    if (n1 > n2)                                // Comparing non-sign words.
    {
      do                                        // Cycle of comparison.
      {
        lp1 = p1 [n1]; --n1;                    // Read word.
        if (lp1 != lp2) goto NotSame;           // Words not the same.
      }
      while (n1 > n2);                          // End of cycle.
    }
  }
  else // (n1 < n2)                             // If number p2 longer, then
  {                                             // compare it with expansion
    lp1 >>= (BITS-1); ++n1;                     // of sign of p2.
    if (lp1 != lp2) goto NotSameSignWord;       // Sign words not the same.
    if (n1 < n2)                                // Comparing non-sign words.
    {
      do                                        // Cycle of comparison.
      {
        lp2 = p2 [n2]; --n2;                    // Read of words.
        if (lp1 != lp2) goto NotSame;           // Words are not the same.
      }
      while (n1 < n2);                          // End of cycle.
    }
  }

//      Pair-wise comparison of non-sign words.

  if (n1)
  {
    do                                          // Cycle of comparison.
    {
      lp1 = p1 [n1]; lp2 = p2 [n1]; --n1;       // Read word.
      if (lp1 != lp2) goto NotSame;             // Words are not the same.
    }
    while (n1);                                 // End of cycle.
  }

  return (0);                                   // Numbers are equal.
NotSame:                                        // Numbers are not equal.
  return (((unsigned long)lp1 >= (unsigned long)lp2) * 2 - 1);
NotSameSignWord:                                // Sign words are not equal.
  return ((lp1 >= lp2) * 2 - 1);
}

//      Special comparison for division.
//      It is assumed:
//      - the numbers are normalized.
//      - size of first number not less then size of second number.
//      Before comparison second number is left shifted to make sizes the same.

int     cBigNumberCompHigh (                    // Comparison from high words.
                const   EXPTR(long) p1,         // Number.
                const   EXPTR(long) p2          // Number.
        )                                       // Returns -1 if p1 <  p2
                                                // Returns  0 if p1 == p2
                                                // Returns  1 if p1  > p2
{                                               // p1, p2 may overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words in p1.
  size_t n2 = (size_t)(*p2);                    // Number of words in p2.
  assert (n1 >= n2);                            // Check of size.
  assert (n2 > 0);                              // Check of size.

  register long lp1 = p1 [n1--];                // Current word of p1.
  register long lp2 = p2 [n2--];                // Current word of p2.
  assert (!n1 || lp1 != (p1 [n1] >> (BITS-1))); // Check of normalization.
  assert (!n2 || lp2 != (p2 [n2] >> (BITS-1))); // Check of normalization.

//      Comparison of sign words.

  if (lp1 != lp2) return ((lp1 > lp2) * 2 - 1); // Words are not the same.

//      Pair-wise comparison of non-sign words.

  if (n2)
  do                                            // Cycle of comparison.
  {
    lp1 = p1 [n1--]; lp2 = p2 [n2--];           // Read words.
    if (lp1 != lp2)                             // Words are not the same.
      return (((unsigned long)lp1 > (unsigned long)lp2) * 2 - 1);
  }
  while (n2);                                   // End of cycle.

//      Comparison over zone of shift.

  if (n1)
  do
  {
    if (p1 [n1--] != 0) return 1;               // Not the same.
  }
  while (n1);

  return (0);                                   // Numbers are equal.
}

//      Special comparison of normalized number with 0.
//
//      Algorithm is optimized to exclude conditional branches.
//      It returns bit-wise sum of sign word and number of words
//      divided by 2 (sign word of positive number may be 0,
//      but in such case number of words > 1).

inline  long    cBigNumberNot0  (               // Comparison with 0:
                const EXPTR(long) p1            // returns 0 if p1 == 0
                )                               // else non 0.
{
  assert (cBigNumberIsFit (p1));                // Check of normalization.
  size_t n1 = (size_t)(*p1);
  return (p1 [n1] | (long)(n1 >> 1));
}

//================================================
//      Copying and bit-wise inversion.
//================================================

#ifndef _CBIGNUM_OPTIMIZED_COPY                 // If not in Cbignumf.cpp.

//      Exact copying.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

void    cBigNumberCopy (                        // Copying p = p1.
                const   EXPTR(long) p1,         // Number to copy.
                        EXPTR(long) p           // Buffer of size *p1 + 1.
        )                                       // p1, p may overlap.
{
  size_t n1 = (size_t)(*p++ = *p1++);           // Number of words.
  if (n1)
  do                                            // Cycle of copying of words.
    { *p++ = *p1++; }                           // Copying of word.
  while (--n1 != 0);                            // End of cycle.
}

#endif//_CBIGNUM_OPTIMIZED_COPY

//      Bit-wise inversion.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

void    cBigNumberCompl (                       // Inversion p = ~p1.
                const   EXPTR(long) p1,         // Number to invert.
                        EXPTR(long) p           // Buffer of size *p1 + 1.
        )                                       // p1, p may overlap.
{
  size_t n1 = (size_t)(*p++ = *p1++);           // Number of words.
  if (n1)
  do                                            // Cycle of inversion of words.
    { *p++ = ~*p1++; }                          // Inversion of word.
  while (--n1 != 0);                            // End of cycle.
}

//================================================
//      Shift to given number of words.
//================================================

//      Left shift increases number of words, if number to shift is not 0,
//      Shift is equivalent to multiplication by 2 in power (BITS * k1).
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

void    cBigNumberCopyShl (                     // Shift p = p1 << (BITS*k1).
                const   EXPTR(long) p1,         // Number to shift.
                        size_t      k1,         // Degree of shift in words.
                        EXPTR(long) p           // Buffer of size *p1 + k1 + 1.
        )                                       // p1, p may overlap.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  if (n1 <= 1 && p1 [n1] == 0) k1 = 0;          // Number 0.
  p1 += n1; *p = (long)(n1 + k1); p += (n1 + k1);
  if (n1)
  do                                            // Cycle of copying of words.
    { *p-- = *p1--; }                           // Copying of word.
  while (--n1 != 0);                            // End of cycle.
  if (k1)
  do                                            // Cycle of clearing of words.
    { *p-- = 0; }                               // Clearing of word.
  while (--k1 != 0);                            // End of cycle.
}

//      Right shift saves sign of number.
//      Shift is equivalent to division by 2 in power (BITS * k1).
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

void    cBigNumberCopyShr (                     // Shift p = p1 >> (BITS*k1).
                const   EXPTR(long) p1,         // Number to shift.
                        size_t      k1,         // Degree of shift in words.
                        EXPTR(long) p           // Buffer of size
                                                // max (*p1 - k1, 1) + 1.
        )                                       // p1, p may overlap.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  if (n1 > k1)                                  // Copying is required.
  {
    *p++ = (long)(n1 -= k1);                    // Number of words.
    p1 += (k1 + 1);                             // Shift.
    do                                          // Cycle of copying of words.
      { *p++ = *p1++; }                         // Copying of word.
    while (--n1 != 0);                          // End of cycle.
  }
  else                                          // Number 0 or -1.
  {
    *p++ = 1; *p = p1 [n1] >> (BITS-1);
  }
}

//      Clipping of some number of words and adding of high 0 word.
//
//      Function returns NON-normalized result.

void    cBigNumberCutOut (                      // Clipping.
                const   EXPTR(long) p1,         // Number to clip.
                        size_t      k1,         // Number of words to skip.
                        size_t      n,          // Number of words to copy.
                        EXPTR(long) p           // Buffer of size n1 + 1.
        )                                       // p1, p may overlap.
{
  assert ((size_t)(*p1) >= k1 + n);             // Check of size.
  *p++ = (long)(n + 1);                         // Number of words.
  if (n > 0)                                    // Copying is required.
  {
    p1 += (k1 + 1);                             // Start.
    do                                          // Cycle of copying of words.
      { *p++ = *p1++; }                         // Copying of word.
    while (--n != 0);                           // End of cycle.
  }
  *p = 0;                                       // The high word.
}

//================================================
//      Unary operations.
//================================================

//      Sign inversion is actually bit-wise inversion followed by increment.
//      1) Invert sign of words until first non-0 word.
//      2) Invert (bit-wise) the successive words.
//      3) If words are over, and the high is 100..(bit-wise),
//         then add word 0.
//         If two high words of result are 11.. and 100..(bit-wise),
//         then remove the high word.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberNeg (                        // Sign inversion p = -p1.
                const   EXPTR(long) p1,         // Initial number.
                        EXPTR(long) p           // Buffer of size *p1 + 2.
        )                                       // p1, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  p++;
  size_t n = n1;                                // Counter of words.
  if (n)
  do                                            // Cycle for sign inversion.
  {
    long lp;                                    // Optimization.
    if ((*p = lp = -*p1) != 0)                  // Sign inversion of word.
    {                                           // Non-0 word obtained.
      if (--n != 0)                             // Are there any words anymore?
      {                                         // If there are:
        do                                      // Cycle of bit-wise inversion.
          { ++p1; ++p; *p = lp = ~*p1; }        // bit-wise inversion of word.
        while (--n != 0);                       // End of cycle of inversion.
        return (n1 - (lp == (p [-1] >> (BITS-1))));
      }                                         // Remove redundant high word.
                                                // If not wore words then
      if (lp != LONG_MIN) return (n1);          // check for overflow
      p [1] = 0; return (n1 + 1);               // and add high word.
    }
  }
  while ((++p1, ++p, --n) != 0);                // End of cycle.

  return (n1);                                  // All words are 0.
}
//#define cBigNumberNeg(p1,p)   (size_t)(*(p)=_cBigNumberNeg(p1,p))

//      Increment:
//      1) Add 1 to words until first non-null result.
//      2) If no more words and the high word is 100..(bit-wise),
//         the add word 0.
//         If two high words of result are 11.. and 100..(bit-wise),
//         then remove the high word.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberMInc (                       // Addition p1 = p1 + 1.
                        EXPTR(long) p1          // Buffer of size *p1 + 2.
        )
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n = n1;                                // Counter of words.

  if (n)                                        // There is at least one word.
  {
    do                                          // Cycle of incrementing.
    {
      long lp;                                  // Optimization.
      if ((*p1 = lp = *p1 + 1) != 0)            // Increment.
      {                                         // Non 0 word obtained.
        if (--n != 0)                           // Are there any words anymore?
        {                                       // If one word is remained and
          if (--n != 0) return (n1);            // it is redundant, delete it.
          return (n1 - (p1 [1] == (lp >> (BITS-1))));
        }                                       // If not wore words then
        if (lp != LONG_MIN) return (n1);        // check for overflow
        p1 [1] = 0; return (n1 + 1);            // and add high word.
      }
    }
    while ((++p1, --n) != 0);                   // End of cycle.
  }
  else { *p1 = 1; n1 = 1; }                     // Normalization of empty num.

  return (n1);                                  // All words are 0.
}
//#define cBigNumberMInc(p1)    (size_t)(*(p1)=_cBigNumberMInc(p1))

//      Decrement:
//      1) Subtract 1 from words until first result, different from ~0.
//      2) If no more words and the high word is 011..(bit-wise),
//         the add word ~0.
//         If two high words of result are 00.. and 011..(bit-wise),
//         then remove the high word.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberMDec (                       // Subtraction p1 = p1 - 1.
                        EXPTR(long) p1          // Buffer of size *p1 + 2.
        )
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n = n1;                                // Counter of words.

  if (n)                                        // There is at least one word.
  {
    do                                          // Cycle of subtracting 1.
    {
      long lp;                                  // Optimization.
      if ((*p1 = lp = *p1 - 1) != ~0)           // Subtraction 1.
      {                                         // Non ~0 word obtained.
        if (--n != 0)                           // Are there any words anymore?
        {                                       // If one word is remained and
          if (--n != 0) return (n1);            // it is redundant, delete it.
          return (n1 - (p1 [1] == (lp >> (BITS-1))));
        }                                       // If no more words then
        if (lp != LONG_MAX) return (n1);        // check for overflow
        p1 [1] = ~0; return (n1 + 1);           // and add high word.
      }
    }
    while ((++p1, --n) != 0);                   // End of cycle.
  }
  else { *p1 = -1; n1 = 1; }                    // Normalization of empty num.

  return (n1);                                  // All words are ~1.
}
//#define cBigNumberMDec(p1)    (size_t)(*(p1)=_cBigNumberMDec(p1))

//      Left shift is performed successively with carry of the high bit
//      to successive word and adding of word with extension of sign bit.
//      If two highest bits of the high word are not the same then number
//      of words is increased - that is, added word becomes sign word
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberMMul2 (                      // Shift p1 <<= 1.
                        EXPTR(long) p1          // Buffer of size *p1 + 2.
        )
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n = n1;                                // Counter of words.
  long num = 0;                                 // Word before shift.
  if (n)
  do                                            // Cycle of shift.
  {
    long pass = (unsigned long)num >> (BITS-1); // Bit from the previous word.
    num = *p1;                                  // Current word before shift.
    *p1++ = (num << 1) | pass;                  // Current word after shift.
  }
  while (--n != 0);                             // End of cycle.

  num >>= (BITS-2);                             // Extension of sign
  n1 += (num != (*p1 = num >> 1));              // on overflow.

  return (n1);                                  // Number of words.
}
//#define cBigNumberMMul2(p1)   (size_t)(*(p1)=_cBigNumberMMul2(p1))

//      Left shift with copying.

size_t  _cBigNumberMul2 (                       // Shift p = p1 << 1.
                const   EXPTR(long) p1,         // Operand.
                        EXPTR(long) p           // Buffer of size *p1 + 2.
        )
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  p++;
  long num = 0;                                 // Word before shift.
  size_t n = n1;                                // Counter of words.

  if (n)
  do                                            // Cycle of shift.
  {
    long pass = (unsigned long)num >> (BITS-1); // Bit from the previous word.
    num = *p1++;                                // Current word before shift.
    *p++ = (num << 1) | pass;                   // Current word after shift.
  }
  while (--n != 0);                             // End of cycle.

  num >>= (BITS-2);                             // Extension of sign
  n1 += (num != (*p = num >> 1));               // on overflow.

  return (n1);                                  // Number of words.
}
//#define cBigNumberMul2(p1,p)  (size_t)(*(p)=_cBigNumberMul2(p1,p))

//      Right shift is performed successively with carry of the low bit
//      to preceding word and adding of word with extension of sign bit.
//      Shift of the high word is signed then unsigned.
//      If before shift all bits if the high word are the same it is deleted.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberMDiv2 (                      // Shift p1 <<= 1.
                        EXPTR(long) p1          // Buffer of size *p1 + 1.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n = n1;                                // Counter of words.
  p1 += n;                                      // The high word.
  if (n)                                        // Is there any words?
  {
    if (--n != 0)                               // More the one word.
    {
      long num = *p1;                           // The high word before shift.
      n1 -= (num == (*p1-- = (num >> 1)));      // Shift of the high word and
                                                // it's deleting of 0 or ~0
      do                                        // Cycle of shift.
      {
        long pass = num << (BITS-1);            // Bit from the next word.
        num = *p1;                              // Current word before shift.
        *p1-- = ((unsigned long)num >> 1) | pass;
      }                                         // Current word after shift.
      while (--n != 0);                         // End of cycle.
    }
    else                                        // Only one word.
    {
      (*p1) >>= 1;                              // Shift.
    }
  }
  return (n1);                                  // Number of words.
}
//#define cBigNumberMDiv2(p1)   (size_t)(*(p1)=_cBigNumberMDiv2(p1))

//      Right shift with copying.

size_t  _cBigNumberDiv2 (                       // Shift p = p1 << 1.
                const   EXPTR(long) p1,         // Operand.
                        EXPTR(long) p           // Buffer of size *p1 + 1.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n = n1;                                // Counter of words.
  p1 += n;                                      // The high word.
  p  += n;                                      // The high word.
  if (n)                                        // Is there any words?
  {
    if (--n != 0)                               // More the one word.
    {
      long num = *p1--;                         // The high word before shift.
      n1 -= (num == (*p-- = (num >> 1)));       // Shift of the high word and
                                                // it's deleting of 0 or ~0
      do                                        // Cycle of shift.
      {
        long pass = num << (BITS-1);            // Bit from the next word.
        num = *p1--;                            // Current word before shift.
        *p-- = ((unsigned long)num >> 1) | pass;// Current word after shift.
      }
      while (--n != 0);                         // End of cycle.
    }
    else                                        // Only one word.
    {
      *p = (*p1) >> 1;                          // Shift.
    }
  }
  return (n1);                                  // Number of words.
}
//#define cBigNumberDiv2(p1,p)  (size_t)(*(p)=_cBigNumberDiv2(p1,p))

//================================================
//      Preparing of table of shifts.
//================================================

//      Number is normalized, then in the same buffer after the number
//      table of shifts by 1..BITS-1 with step *p + 2 is prepared.
//      Function always returns normalized number and normalized shifts;
//      non-involved high words is filled by extensions of sign bits.
//      Number with table of shifts can take place in optimized
//      operations that use table of shifts.

void    cBigNumberTab  (                        // Preparing table of shifts.
                        EXPTR(long) p           // Number in buffer of size
        )                                       // (*p + 2) * BITS + 1.
                                                // Buffer is filled by shifts
{                                               // p << 1..BITS-1 step *p + 2.
  size_t n = cBigNumberFit (p);                 // Number of normalized words.
  int ns = BITS - 1;                            // Counter of bit shifts.
  p [n + 1] = p [n] >> (BITS-1);                // Extension of sign.
  n += 2;
  EXPTR(long) pp = p;                           // Work pointer.
  do                                            // Cycle of shifts.
  {
    pp += n;
    cBigNumberMul2 (p, pp);                     // Left bit shift of number.
    p = pp;
  }
  while (--ns > 0);                             // End of cycle.
}

//================================================
//      Logical functions.
//================================================

//      Sign bit of more short number expanded to make
//      number of words in both numbers the same.
//      Then the numbers are processed word by word.

//      Peculiarities:
//      - The second operand must not be longer then the first one.
//      - Functions always returns normalized number.
//      - For optimization purposes algorithmic equivalent
//        is applied instead of actual expanding of sign bit.
//
//      Functions do not require for normalization of operands and
//      normalize the result, except for case when operand contains
//      0 words - in such case the result will also contain 0 words.

size_t  _cBigNumberXor  (                       // Addition mod2 p = p1 ^ p2.
                const   EXPTR(long) p1,         // Operand.
                const   EXPTR(long) p2,         // Operand (no longer p1).
                        EXPTR(long) p           // Buffer of size *p1 + 1.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n2 = (size_t)(*p2++);                  // Number of words.
  p++;
  assert (n1 >= n2);
  if (n1 == 0) return 0;                        // Return if 0 words.

  {
    register long lp2 = 0;                      // Current word of p2.

//      Processing of words p1, p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ ^ (lp2 = *p2++); }    // Cycle on p1, p2.
        while (--n != 0);                       // End of cycle on p1, p2.
      }
    }

//      Processing of remainder p1, if p1 is longer then p2.

    {
      lp2 >>= (BITS-1);                         // Sign bit of p2.
      size_t n;                                 // Counter of words.
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ ^ lp2; }              // Cycle on p1.
        while (--n != 0);                       // End of cycle on p1.
      }
    }
    --p;                                        // Preparing of pointer.
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.

  {
    register long pass;                         // The high word.
    register long lp;                           // The previous word.
    register size_t n = n1;                     // Number of words.
    if (n)                                      // If bits of the high word
    {                                           // and high bit of the word
      lp = *p;                                  // before high are all 0 or 1
      do continue;                              // then delete the high word.
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp >> (BITS-1))));
    }                                           // If no words at all
    else { p [1] = 0; }                         // then add word 0.

    return (n + 1);                             // Number of words p.
  }
}
//#define cBigNumberXor(p1,p2,p) (size_t)(*(p)=_cBigNumberXor(p1,p2,p))

size_t  _cBigNumberAnd  (                       // Conjunction p = p1 & p2.
                const   EXPTR(long) p1,         // Operand.
                const   EXPTR(long) p2,         // Operand (no longer p1).
                        EXPTR(long) p           // Buffer of size *p1 + 1.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n2 = (size_t)(*p2++);                  // Number of words.
  p++;
  assert (n1 >= n2);
  if (n1 == 0) return 0;                        // Return if 0 words.

  {
    register long lp2 = 0;                      // Current word of p2.

//      Processing of words p1, p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ & (lp2 = *p2++); }    // Cycle on p1, p2.
        while (--n != 0);                       // End of cycle on p1, p2.
      }
    }

//      Processing of remainder p1, if p1 is longer then p2.

    {
      lp2 >>= (BITS-1);                         // Sign bit of p2.
      size_t n;                                 // Counter of words.
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ & lp2; }              // Cycle on p1.
        while (--n != 0);                       // End of cycle on p1.
      }
    }
    --p;                                        // Preparing of pointer.
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.

  {
    register long pass;                         // The high word.
    register long lp;                           // The previous word.
    register size_t n = n1;                     // Number of words.
    if (n)                                      // If bits of the high word
    {                                           // and high bit of the word
      lp = *p;                                  // before high are all 0 or 1
      do continue;                              // then delete the high word.
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp >> (BITS-1))));
    }                                           // If no words at all
    else { p [1] = 0; }                         // then add word 0.

    return (n + 1);                             // Number of words p.
  }
}
//#define cBigNumberAnd(p1,p2,p) (size_t)(*(p)=_cBigNumberAnd(p1,p2,p))

size_t  _cBigNumberOr  (                        // Disjunction p = p1 & p2.
                const   EXPTR(long) p1,         // Operand.
                const   EXPTR(long) p2,         // Operand (no longer p1).
                        EXPTR(long) p           // Buffer of size *p1 + 1.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n2 = (size_t)(*p2++);                  // Number of words.
  p++;
  assert (n1 >= n2);
  if (n1 == 0) return 0;                        // Return if 0 words.

  {
    register long lp2 = 0;                      // Current word of p2.

//      Processing of words p1, p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ | (lp2 = *p2++); }    // Cycle on p1, p2.
        while (--n != 0);                       // End of cycle on p1, p2.
      }
    }

//      Processing of remainder p1, if p1 is longer then p2.

    {
      lp2 >>= (BITS-1);                         // Sign bit of p2.
      size_t n;                                 // Counter of words.
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ | lp2; }              // Cycle on p1.
        while (--n != 0);                       // End of cycle on p1.
      }
    }
    --p;                                        // Preparing of pointer.
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.

  {
    register long pass;                         // The high word.
    register long lp;                           // The previous word.
    register size_t n = n1;                     // Number of words.
    if (n)                                      // If bits of the high word
    {                                           // and high bit of the word
      lp = *p;                                  // before high are all 0 or 1
      do continue;                              // then delete the high word.
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp >> (BITS-1))));
    }                                           // If no words at all
    else { p [1] = 0; }                         // then add word 0.

    return (n + 1);                             // Number of words p.
  }
}
//#define cBigNumberOr(p1,p2,p)  (size_t)(*(p)=_cBigNumberOr(p1,p2,p))

//================================================
//      Functions of addition and subtraction.
//================================================

//      Sign bit of more short number expanded to make
//      number of words in both numbers the same.
//      Then the numbers are added word by word with carrying.
//
//      Peculiarities:
//      - The second operand must not be longer then the first one.
//      - Function always returns normalized number.
//      - For optimization purposes algorithmic equivalent
//        is applied instead of actual expanding of sign bit.

size_t  _cBigNumberAdd  (                       // Addition p = p1 + p2.
                const   EXPTR(long) p1,         // Augend.
                const   EXPTR(long) p2,         // Addend (no longer p1).
                        EXPTR(long) p           // Buffer of size *p1 + 2.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n2 = (size_t)(*p2++);                  // Number of words.
  p++;
  assert (n1 >= n2);
  if (n1 == 0) { --p1; ++n1; }                  // Correction for 0.

  register unsigned long pass = 0;              // Carry to next word of p.
        // carry is contained in the high bit of pass
        // unsigned long provides for unsigned shift of pass.
  {
    register long lp1 = 0;                      // Current word of p1.
    register long lp2 = 0;                      // Current word of p2.

//      Addition of words p1, p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n2) != 0)
      {
        do                                      // Cycle on p1, p2.
        {
          lp1 = *p1++; lp2 = *p2++;
          pass = (~(*p++ = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                           | (lp1 & lp2);
        }
        while (--n != 0);                       // End of cycle on p1, p2.
      }
    }

//      Either carry or borrow over rest of p1 if p1 is longer then p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;           // Optimization.
        if ((lp2 & LONG_MIN) == 0)              // p2 is not negative.
        {
          do                                    // Cycle on p1.
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);                     // End of cycle on p1.
        }
        else                                    // p2 is negative.
        {
          do                                    // Cycle on p1.
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);                     // End of cycle on p1.
        }
      }
    }
    --p;                                        // Preparing of pointer.

//      Addition of sign extensions of p1, p2.
//      This produced the high word of p consists of solely 0 or 1
//      The high word p is stored in pass.

    pass = (pass >> (BITS-1)) + ((lp1) >> (BITS-1)) + ((lp2) >> (BITS-1));
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.

  {
    register long lp = *p;                      // The previous word.
    register size_t n = n1;                     // Number of words - 1.
    if (n && (long)pass == (lp >> (BITS-1)))    // If bits of the high word
    {                                           // and high bit of the word
      do continue;                              // before high are all 0 or 1
      while (--n != 0 &&                        // then delete the high word.
        (pass = lp, lp = p [-1], --p, (long)pass == (lp >> (BITS-1))));
    }
    else { p [1] = pass; }                      // Storing of the high word.

    return (n + 1);                             // Number of words p.
  }
}
//#define cBigNumberAdd(p1,p2,p) (size_t)(*(p)=_cBigNumberAdd(p1,p2,p))

//      Sign bit of more short number expanded to make
//      number of words in both numbers the same.
//      Then the numbers are subtracted word by word with borrowing.
//
//      Peculiarities:
//      - The second operand must not be longer then the first one.
//      - Functions always returns normalized number.
//      - For optimization purposes algorithmic equivalent
//        is applied instead of actual expanding of sign bit.
//      - Subtraction is implemented in two variants:
//        conventional and with sign inversion.

size_t  _cBigNumberSub  (                       // Subtraction p = p1 - p2.
                const   EXPTR(long) p1,         // Subtrahend.
                const   EXPTR(long) p2,         // Subtracter (no longer p1).
                        EXPTR(long) p           // Buffer of size *p1 + 2.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n2 = (size_t)(*p2++);                  // Number of words.
  p++;
  assert (n1 >= n2);
  if (n1 == 0) { --p1; ++n1; }                  // Correction for 0.

  register unsigned long pass = ~0UL;           // Carry to next word of p.
        // carry is contained in the high bit of pass
        // unsigned long provides for unsigned shift of pass.
        // initial value provides for addition of 1.
  {
    register long lp1 = 0;                      // Current word of p1.
    register long lp2 = ~0;                     // Current word of ~p2.

//      Addition of words p1, ~p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n2) != 0)
      {
        do                                      // Cycle on p1, p2.
        {
          lp1 = *p1++; lp2 = ~(*p2++);
          pass = (~(*p++ = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                           | (lp1 & lp2);
        }
        while (--n != 0);                       // End of cycle on p1, p2.
      }
    }

//      Either carry or borrow over rest of p1 if p1 is longer then p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;           // Optimization.
        if ((lp2 & LONG_MIN) == 0)              // p2 is negative.
        {
          do                                    // Cycle on p1.
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);                     // End of cycle on p1.
        }
        else                                    // p2 is not negative.
        {
          do                                    // Cycle on p1.
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);                     // End of cycle on p1.
        }
      }
    }
    --p;                                        // Preparing of pointer.

//      Addition of sign extensions of p1, p2~.
//      This produced the high word of p consists of solely 0 or 1
//      The high word p is stored in pass.

    pass = (pass >> (BITS-1)) + ((lp1) >> (BITS-1)) + ((lp2) >> (BITS-1));
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.

  {
    register long lp = *p;                      // The previous word.
    register size_t n = n1;                     // Number of words - 1.
    if (n && (long)pass == (lp >> (BITS-1)))    // If bits of the high word
    {                                           // and high bit of the word
      do continue;                              // before high are all 0 or 1
      while (--n != 0 &&                        // then delete the high word.
        (pass = lp, lp = p [-1], --p, (long)pass == (lp >> (BITS-1))));
    }
    else { p [1] = pass; }                      // Storing of the high word.

    return (n + 1);                             // Number of words p.
  }
}
//#define cBigNumberSub(p1,p2,p) (size_t)(*(p)=_cBigNumberSub(p1,p2,p))

//      Subtraction with sign inversion.

size_t  _cBigNumberSubS (                       // Subtraction p = p2 - p1.
                const   EXPTR(long) p1,         // Subtracter.
                const   EXPTR(long) p2,         // Subtrahend (no longer p1).
                        EXPTR(long) p           // Buffer of size *p1 + 2.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1++);                  // Number of words.
  size_t n2 = (size_t)(*p2++);                  // Number of words.
  p++;
  assert (n1 >= n2);
  if (n1 == 0) { --p1; ++n1; }                  // Correction for 0.

  register unsigned long pass = ~0UL;           // Carry to next word of p.
        // carry is contained in the high bit of pass
        // unsigned long provides for unsigned shift of pass.
        // initial value provides for addition of 1.
  {
    register long lp1 = ~0;                     // Current word of ~p1.
    register long lp2 = 0;                      // Current word of p2.

//      Addition of words ~p1, p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n2) != 0)
      {
        do                                      // Cycle on p1, p2.
        {
          lp1 = ~(*p1++); lp2 = *p2++;
          pass = (~(*p++ = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                           | (lp1 & lp2);
        }
        while (--n != 0);                       // End of cycle on p1, p2.
      }
    }

//      Either carry or borrow over rest of p1 if p1 is longer then p2.

    {
      size_t n;                                 // Counter of words.
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;           // Optimization.
        if ((lp2 & LONG_MIN) == 0)              // p2 is negative.
        {
          do                                    // Cycle on p1.
          {
            lp1 = ~(*pp++);
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);                     // End of cycle on p1.
        }
        else                                    // p2 is not negative.
        {
          do                                    // Cycle on p1.
          {
            lp1 = ~(*pp++);
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);                     // End of cycle on p1.
        }
      }
    }
    --p;                                        // Preparing of pointer.

//      Addition of sign extensions of ~p1, p2.
//      This produced the high word of p consists of solely 0 or 1
//      The high word p is stored in pass.

    pass = (pass >> (BITS-1)) + ((lp1) >> (BITS-1)) + ((lp2) >> (BITS-1));
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.

  {
    register long lp = *p;                      // The previous word.
    register size_t n = n1;                     // Number of words - 1.
    if (n && (long)pass == (lp >> (BITS-1)))    // If bits of the high word
    {                                           // and high bit of the word
      do continue;                              // before high are all 0 or 1
      while (--n != 0 &&                        // then delete the high word.
        (pass = lp, lp = p [-1], --p, (long)pass == (lp >> (BITS-1))));
    }
    else { p [1] = pass; }                      // Storing of the high word.

    return (n + 1);                             // Number of words p.
  }
}
//#define cBigNumberSubS(p1,p2,p) (size_t)(*(p)=_cBigNumberSubS(p1,p2,p))

//================================================
//      Special fast functions of accumulation.
//================================================

//      Function to count number of low 0-words.

static  size_t  _cBigNumberSkip = 1;            // Number of low 0-words + 1.

size_t  cBigNumberSkipLow0 (                    // Count low 0-words.
                const   EXPTR(long) p1          // Buffer with number.
)                                               // Returns num of 0-words + 1.
{
  size_t n1 = (size_t)(*p1);                    // Number of words p1.
  size_t n = 1;                                 // Number of low 0-words + 1.

  if (n1 != 0)
  do
  {
    if (p1 [n] != 0) break;                     // Non-0 word is found.
    n++;
  }
  while (--n1 != 0);

#ifndef _CBIGNUM_MT
  _cBigNumberSkip = n;                          // Number of low 0-words + 1.
#endif//_CBIGNUM_MT

  return (n);                                   // Number of low 0-words + 1.
}

#ifndef _CBIGNUM_ASM                            // If not in Cbignumf.cpp.

//      Special addition with left shift for multiplication.
//
//      For optimization purposes it is assumed that:
//      - Size of the first number is not smaller then size of the second
//        number after shift and not smaller then size of result.
//      - Size of the second number before shift is greater then 0.
//      Before addition one must call function cBigNumberSkipLow0()
//      to count number of low 0-words in the second operand.
//
//      Function does not require for normalization of operands, does not
//      provides for normalization of result and does not change size of
//      the first number.

void    cBigNumberMAddM (                       // Addition p1 += p2<<k2*BITS.
                        EXPTR(long) p1,         // Augend, then result.
                const   EXPTR(long) p2,         // Addend of size <=*p1-k2.
                        size_t      k2          // Left shift of addend.
        )                                       // p1, p2 may overlap.
{
  const EXPTR(long) pp = p1 + (size_t)(*p1);    // The last word of p1.
  size_t n2 = (size_t)(*p2);                    // Number of words p2.
  assert (_cBigNumberSkip > 0);                 // Check of parameter.
  assert ((size_t)(*p1) >= n2 + k2);            // Check of size.
  assert (n2 >= _cBigNumberSkip);               // Check of size.
  p1 += k2;                                     // Scaling of p2.

  register unsigned long pass = 0;              // Carry: 0 or 1.
        // unsigned long provides for unsigned shift of pass.
  {
    const EXPTR(long) pp = p1 + n2;             // The last word of p2.
    p1 += _cBigNumberSkip;                      // Set start of p1.
    p2 += _cBigNumberSkip;                      // Set start of p2.
    register long lp1;                          // Current word of p1.
    register long lp2;                          // Current word of p2.

//      Addition of words p1, p2.
//      Operators of cycle calculates the following logical expression:
//      pass = ((~(*p1 = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
//                                                       | (lp1 & lp2))
//      Logical expression is listed by the following chain of assignments:
//      pass += (lp1 + lp2)     // *p1 = pass
//      lp1  = ~(lp1 ^ lp2)     // Assignment to lp1 releases one register.
//      pass = ~(pass | lp1)    // It is used identity of de Morgan.
//      pass |= (lp1 & lp2)     // and identity a & b = ~(a^b) & b.
//      Assignments are interleaved by cycle handling checks and increments
//      of pointers to optimize loading of Pentium pipes.

    for (;;)                                    // Cycle on p1, p2.
    {
      lp1 = *p1;        lp2 = *p2;              // Current words of p1, p2.
      pass += lp1;      lp1 ^= lp2;
      pass += lp2;      lp1 = ~lp1;
      *p1 = pass;       pass |= lp1;
      lp1 &= lp2;       pass = ~pass;
      ++p2;             pass |= lp1;
      if (p1 >= pp)     break;
      ++p1;             pass >>= (BITS-1);
    }
    if ((long)(lp2 ^= pass) >= 0) return;       // No carry.
  }

//      Either carry or borrow over rest of p1 if p1 is longer then p2.
//      Starting condition: p1 points to last word obtained,
//                          pass contains carry in the high bit.

  {
    register long lp1;                          // Current word of p1.
    if ((long)pass < 0)                         // p2 is not negative.
    {
      do                                        // Cycle on p1.
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 + 1) & lp1) < 0);
    }
    else                                        // p2 is negative.
    {
      do                                        // Cycle on p1.
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 - 1) | lp1) >= 0);
    }
  }
}

//      Special subtraction with left shift for multiplication.
//
//      For optimization purposes it is assumed that:
//      - Size of the first number is not smaller then size of the second
//        number after shift and not smaller then size of result.
//      - Size of the second number before shift is greater then 0.
//      Before subtraction one must call function cBigNumberSkipLow0()
//      to count number of low 0-words in the second operand.
//
//      Function does not require for normalization of operands, does not
//      provides for normalization of result and does not change size of
//      the first number.

void    cBigNumberMSubM (                       // Subtract p1 -= p2<<k2*BITS.
                        EXPTR(long) p1,         // Subtrahend, then result.
                const   EXPTR(long) p2,         // Subtracter of size <=*p1-k2.
                        size_t      k2          // Left shift of subtracter.
        )                                       // p1, p2 may overlap.
{
  const EXPTR(long) pp = p1 + (size_t)(*p1);    // The last word of p1.
  size_t n2 = (size_t)(*p2);                    // Number of words p2.
  assert (_cBigNumberSkip > 0);                 // Check of parameter.
  assert ((size_t)(*p1) >= n2 + k2);            // Check of size.
  assert (n2 >= _cBigNumberSkip);               // Check of size.
  p1 += k2;                                     // Scaling of p2.

  register long pass = 0;                       // Borrow: 0 or -1.
        // long provides for signed shift of pass.
  {
    const EXPTR(long) pp = p1 + n2;             // The last word of p2.
    p1 += _cBigNumberSkip;                      // Set start of p1.
    p2 += _cBigNumberSkip;                      // Set start of p2.
    register long lp1;                          // Current word of p1.
    register long lp2;                          // Current word of p2.

//      Subtraction of words p1, p2.
//      Operators of cycle calculates the following logical expression:
//      pass = (((*p1 = (pass >> (BITS-1)) + lp1 - lp2) | (lp1 ^ lp2))
//                                                     & (~lp1 | lp2))
//      Logical expression is listed by the following chain of assignments:
//      pass += (lp1 - lp2)     // *p1 = pass
//      pass |= (lp1 ^= lp2)    // Assignment to lp1 releases one register.
//      pass = ~pass;           // It is used identity of de Morgan
//      pass |= (lp1 & ~lp2)    // and identity a & ~b = (a^b) & ~b.
//      pass = ~pass;
//      Assignments are interleaved by cycle handling checks and increments
//      of pointers to optimize loading of Pentium pipes.

    for (;;)                                    // Cycle on p1, p2.
    {
      lp1 = *p1;        lp2 = *p2;              // Current words of p1, p2.
      pass += lp1;      lp1 ^= lp2;
      pass -= lp2;      lp2 = ~lp2;
      *p1 = pass;       pass |= lp1;
      lp1 &= lp2;       pass = ~pass;
      ++p2;             pass |= lp1;
      if (p1 >= pp)     break;
      ++p1;             pass >>= (BITS-1);
                        pass = ~pass;
    }
    if ((long)(lp2 ^= pass) >= 0) return;       // No carry or borrow.
  }

//      Either carry or borrow over rest of p1 if p1 is longer then p2.
//      Starting condition: p1 points to last word obtained,
//                          pass contains ~borrow in the high bit.

  {
    register long lp1;                          // Current word of p1.
    if ((long)pass < 0)                         // p2 is negative.
    {
      do                                        // Cycle on p1.
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 + 1) & lp1) < 0);
    }
    else                                        // p2 is not negative.
    {
      do                                        // Cycle on p1.
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 - 1) | lp1) >= 0);
    }
  }
}

//      Special subtraction with left shift for division.
//
//      For optimization purposes it is assumed that:
//      - Numbers have the same signs.
//      - Size of the first number is not smaller then size of the second
//        number after shift.
//      - Size of the second number before shift is greater then 0.
//      - After subtraction number of meaning bits decreases.
//      - After subtraction sign does not change except for case
//        of zeroing of negative numbers.
//      Before subtraction one must call function cBigNumberSkipLow0()
//      to count number of low 0-words in the subtracter.
//
//      Function does not require for normalization of operands and
//      always returns normalized number.

size_t  _cBigNumberMSubD (                      // Subtract p1 -= p2<<k2*BITS.
                        EXPTR(long) p1,         // Subtrahend, then result.
                const   EXPTR(long) p2,         // Subtracter of size <=*p1-k2.
                        size_t      k2          // Left shift of subtracter.
        )                                       // p1, p2 may overlap.
{
  size_t n2 = (size_t)(*p2);                    // Number of words p2.
  assert (_cBigNumberSkip > 0);                 // Check of parameter.
  assert ((p1 [(size_t)(*p1)] ^ p2 [n2]) >= 0); // Check of sign.
  assert ((size_t)(*p1) >= n2 + k2);            // Check of size.
  assert (n2 >= _cBigNumberSkip);               // Check of size.
  p1 += k2;                                     // Scaling of p2.

  register long pass = 0;                       // Borrow: 0 or -1.
        // long provides for signed shift of pass.
  {
    const EXPTR(long) pp = p1 + n2;             // The last word of p2.
    p1 += _cBigNumberSkip;                      // Set start of p1.
    p2 += _cBigNumberSkip;                      // Set start of p2.
    register long lp1;                          // Current word of p1.
    register long lp2;                          // Current word of p2.

//      Subtraction of words p1, p2.
//      Operators of cycle calculates the following logical expression:
//      pass = (((*p1 = (pass >> (BITS-1)) + lp1 - lp2) | (lp1 ^ lp2))
//                                                     & (~lp1 | lp2))
//      Logical expression is listed by the following chain of assignments:
//      pass += (lp1 - lp2)     // *p1 = pass
//      pass |= (lp1 ^= lp2)    // Assignment to lp1 releases one register.
//      pass &=~(lp1 & ~lp2)    // It is used identity of de Morgan
//                              // and identity a & ~b = (a^b) & ~b.
//      Assignments are interleaved by cycle handling checks and increments
//      of pointers to optimize loading of Pentium pipes.

    for (;;)                                    // Cycle on p1, p2.
    {
      lp1 = *p1;        lp2 = *p2;              // Current words of p1, p2.
      pass += lp1;      lp1 ^= lp2;
      pass -= lp2;      lp2 = ~lp2;
      *p1 = pass;       lp2 &= lp1;
      if (p1 >= pp)     break;
      lp2 = ~lp2;       pass |= lp1;
      ++p1;             pass &= lp2;
      ++p2;             pass >>= (BITS-1);
    }
  }

//      Normalization - deleting of redundant high words of p consisting of
//      0 or 1 and coincident with expansion of sign bit of the previous word.
//      Starting condition: p1 points to last word obtained,
//                          copy of which resides in pass.

  {
    register size_t n1 = n2 + k2;

//      Loop performs the following operations:
//      do continue;
//      while (--n1 != 0 &&
//              lp1 == (lp1 = pass, pass = p1 [-1], --p1, pass >> (BITS-1)));

    for (;;)
    {
      register long lp1;                        // The high word.
      register long lp2;                        // The previous word.
      if (--n1 == 0)    break;                  // One word remains.
      lp2 = p1 [-1];    --p1;
      lp1 = pass;       pass = lp2;
      lp2 >>= (BITS-1);
      if (lp1 != lp2)   break;
    }
    return (n1 + 1);                            // Number of words p.
  }
}
//#define cBigNumberMSubD(p1,p2,k2) (size_t)(*(p1)=_cBigNumberMSubD(p1,p2,k2))

//      Special 1-bit left shift for multiplication.
//
//      It is assumed that the number is not 0.
//      Before shift one must call function cBigNumberSkipLow0()
//      to count number of low 0-words in the buffer.
//
//      Left shift is performed successively with carry of the high bit
//      to successive word and adding of word with extension of sign bit.
//      If two highest bits of the high word are not the same then number
//      of words is increased - that is, added word becomes sign word.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberMMul2M (                     // Shift p1 <<= 1.
                        EXPTR(long) p1          // Buffer of size *p1 + 2.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  assert (_cBigNumberSkip > 0);                 // Check of parameter.
  assert (n1 >= _cBigNumberSkip);               // Check of size.

  const EXPTR(long) pp = p1 + n1;               // The last word of p1.
  p1 += _cBigNumberSkip;                        // First non-0 word of p1.
  long num = 0;                                 // Word before shift.
  do                                            // Cycle of shift.
  {
    long pass = (unsigned long)num >> (BITS-1); // Bit from the previous word.
    num = *p1;                                  // Current word before shift.
    *p1++ = (num << 1) | pass;                  // Current word after shift.
  }
  while (p1 <= pp);                             // Cycle of shift.

  num >>= (BITS-2);                             // Extension of sign
  n1 += (num != (*p1 = num >> 1));              // on overflow.

  return (n1);                                  // Number of words.
}
//#define cBigNumberMMul2M(p1)   (size_t)(*(p1)=_cBigNumberMMul2M(p1))

//      Special 1-bit right shift for division and square root.
//
//      It is assumed:
//      - Number is not 0.
//      - The lower bit of the lower non-0 word is 0.
//      Before shift one must call function cBigNumberSkipLow0()
//      to count number of low 0-words in the buffer.
//
//      Right shift is performed successively with carry of the low bit
//      to preceding word and adding of word with extension of sign bit.
//      Shift of the high word is signed then unsigned.
//      If before shift all bits if the high word are the same it is deleted.
//
//      Function neither requires for normalization of operand
//      nor normalizes it, but returns normalized result if
//      operand is normalized.

size_t  _cBigNumberMDiv2D (                     // Shift p1 >>= 1.
                        EXPTR(long) p1          // Buffer of size *p1 + 1.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  assert (_cBigNumberSkip > 0);                 // Check of parameter.
  assert (n1 >= _cBigNumberSkip);               // Check of size.
  assert ((p1 [_cBigNumberSkip] & 1L) == 0);    // Check of lowest bit.

  const EXPTR(long) pp = p1 + _cBigNumberSkip;  // First non-0 word of p1.
  p1 += n1;                                     // The high word.
  long num = *p1;                               // The high word before shift.
  n1 -= (n1 > 1) & (num == (*p1-- = num >> 1)); // Shift of the high word and
                                                // deleting, if 0 or ~0.
  while (p1 >= pp)                              // Cycle of shift.
  {
    long pass = num << (BITS-1);                // Bit from the next word.
    num = *p1;                                  // Current word before shift.
    *p1-- = ((unsigned long)num >> 1) | pass;   // Current word after shift.
  }                                             // End of cycle.

  return (n1);                                  // Number of words.
}
//#define cBigNumberMDiv2D(p1)   (size_t)(*(p1)=_cBigNumberMDiv2D(p1))

#endif//_CBIGNUM_ASM

//      Special 1-word right shift for multiplication.
//      Initial number must contain at least two words.

void    cBigNumberMShrM (                       // Shift p1 >>= BITS.
                        EXPTR(long) p1          // Buffer of size *p1 + 1.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  assert (n1 > 1);                              // Check.
  *p1 = (long)--n1;
  do                                            // Cycle of copying of words.
    { p1 [1] = p1 [2]; p1++; }                  // Copying of word.
  while ((long)--n1 > 0);                       // End of cycle.
}

//      Special 1-word left shift for division.
//      Initial number must contain at least one word.

void    cBigNumberMShlD (                       // Shift p1 <<= BITS.
                        EXPTR(long) p1          // Buffer of size *p1 + 2.
        )
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  assert (n1 > 0);                              // Check.
  *p1 = (long)(n1 + 1);
  do                                            // Cycle of copying of words.
    { p1 [n1 + 1] = p1 [n1]; }                  // Copying of word.
  while ((long)--n1 > 0);                       // End of cycle.
  p1 [1] = 0;
}

//      Functions of fast accumulation with shift optimized
//      for large numbers, in particular long accumulator.
//
//      Functions does not require for normalization of operands
//      and always return normalized result.

void    cBigNumberMAddShl (                     // Addition p1 += p2<<k2*BITS.
                        EXPTR(long) p1,         // Augend, then result
                                                // in buffer of size
                                                // max (*p1, *p2 + k2) + 2.
                const   EXPTR(long) p2,         // Addend.
                        size_t      k2          // Left shift of addend.
        )                                       // p1, p2 may overlap.
{
  size_t n2 = (size_t)(*p2);                    // Number of words.

  cBigNumberFitTo (p1, n2 + k2 + 1);            // Denormalization.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
  if (n2 > 0) cBigNumberMAddM (p1, p2, k2);     // Fast addition.

  cBigNumberFit (p1);                           // Normalization.
}

void    cBigNumberMSubShl (                     // Subtract p1 -= p2<<k2*BITS.
                        EXPTR(long) p1,         // Subtrahend, then result
                                                // in buffer of size
                                                // max (*p1, *p2 + k2) + 2.
                const   EXPTR(long) p2,         // Subtracter.
                        size_t      k2          // Left shift of subtracter.
        )                                       // p1, p2 may overlap.
{
  size_t n2 = (size_t)(*p2);                    // Number of words.

  cBigNumberFitTo (p1, n2 + k2 + 1);            // Denormalization.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
  if (n2 > 0) cBigNumberMSubM (p1, p2, k2);     // Fast subtraction.

  cBigNumberFit (p1);                           // Normalization.
}

//================================================
//      Functions of multiplication.
//================================================

//      Here is portable binary implementation on standard C++.
//      Look also Cbignumf.cpp for much faster implementation
//      based on hardware multiplication.

#ifndef _CBIGNUM_HARDWARE_MUL                   // If not in Cbignumf.cpp.

//      Special functions for accumulation of multiplication with shift.
//      Multiplicand is used as work buffer and will be overwritten.
//
//      Multiplier is considered to be UNSIGNED number,
//      i.e. sign bits are interpreted as high bits.
//
//      Functions does not require for normalization of operands
//      and always return normalized result.

void    cBigNumberMAddMulShl (                  // Addition of multiplication
                                                // p += p1 * p2 << k*BITS.
                        EXPTR(long) p1,         // Multiplicand in overwritten
                                                // buffer of size *p1 + 3.
                const   EXPTR(long) p2,         // Unsigned multiplier.
                        size_t      k,          // Left shift of product.
                        EXPTR(long) p           // Augend in buffer of size
                                                // max (*p, *p1 + *p2 + k) + 2.
        )                                       // p1, p2, p may not overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  assert (p2 != p1);                            // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

//      Algorithm that does not require for table of shifts.

  if (n1 >= cBigNumberSkipLow0 (p1) && n2 != 0) // Are operands not empty?
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);       // Denormalization.
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)                      // Cycle on words of p2.
    {
      long num = p2 [k2 + 1];                   // Current word of p2.
      if (num == 0) continue;                   // Optimization.
      long mask = 1;                            // Mask for bits of num.
      for (;; cBigNumberMMul2M (p1))            // Cycle on bits of word of p2.
      {
        if (num & mask) cBigNumberMAddM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;           // No more bits.
      }                                         // End of cycle on bits of p2.
      cBigNumberMMul2M (p1);                    // Bit shift of multiplicand.
      cBigNumberMShrM (p1);                     // Reset of multiplicand.
    }                                           // End of cycle on words of p2.
    {
      long num = p2 [k2 + 1];                   // The last word of p2.
      long mask = (num != 0);                   // Mask for bits of num.
      if (num > 0)                              // The high bit of num
      do                                        // may be not 0.
        mask <<= 1;                             // Shift mask and num, while
      while ((num <<= 1) > 0);                  // the high bit of num is 0.
      for (;; cBigNumberMMul2M (p1))            // Cycle on bits of word of p2.
      {
        if (num & mask) cBigNumberMAddM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;           // No more bits.
      }                                         // End of cycle on bits of p2.
    }
  }

  cBigNumberFit (p);                            // Normalization.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

void    cBigNumberMSubMulShl (                  // Subtract of multiplication
                                                // p -= p1 * p2 << k*BITS.
                        EXPTR(long) p1,         // Multiplicand in overwritten
                                                // buffer of size *p1 + 3.
                const   EXPTR(long) p2,         // Unsigned multiplier.
                        size_t      k,          // Left shift of product.
                        EXPTR(long) p           // Subtrahend in buffer of size
                                                // max (*p, *p1 + *p2 + k) + 2.
        )                                       // p1, p2, p may not overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  assert (p2 != p1);                            // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

//      Algorithm that does not require for table of shifts.

  if (n1 >= cBigNumberSkipLow0 (p1) && n2 != 0) // Are operands not empty?
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);       // Denormalization.
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)                      // Cycle on words of p2.
    {
      long num = p2 [k2 + 1];                   // Current word of p2.
      if (num == 0) continue;                   // Optimization.
      long mask = 1;                            // Mask for bits of num.
      for (;; cBigNumberMMul2M (p1))            // Cycle on bits of word of p2.
      {
        if (num & mask) cBigNumberMSubM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;           // No more bits.
      }                                         // End of cycle on bits of p2.
      cBigNumberMMul2M (p1);                    // Bit shift of multiplicand.
      cBigNumberMShrM (p1);                     // Reset of multiplicand.
    }                                           // End of cycle on words of p2.
    {
      long num = p2 [k2 + 1];                   // The last word of p2.
      long mask = (num != 0);                   // Mask for bits of num.
      if (num > 0)                              // The high bit of num
      do                                        // may be not 0.
        mask <<= 1;                             // Shift mask and num, while
      while ((num <<= 1) > 0);                  // the high bit of num is 0.
      for (;; cBigNumberMMul2M (p1))            // Cycle on bits of word of p2.
      {
        if (num & mask) cBigNumberMSubM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;           // No more bits.
      }                                         // End of cycle on bits of p2.
    }
  }

  cBigNumberFit (p);                            // Normalization.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

#endif//_CBIGNUM_HARDWARE_MUL

//      Special functions for accumulation of multiplication with shift.
//      It is assumed that multiplicand contains table of shifts, obtained
//      by cBigNumberTab(). Method is effective if size of at lease one
//      operand is not greater then _CBNL_KARATSUBA_MIN * 2.
//      For better performance larger operand should be placed first,
//      except for case when it size is greater then _CBNL_TAB_OPT.
//
//      Multiplier is considered to be UNSIGNED number,
//      i.e. sign bits are interpreted as high bits.
//
//      Functions does not require for normalization of operands
//      and always return normalized result.

void    cBigNumberMAddMulShlTab (               // Addition of multiplication
                                                // p += p1 * p2 << k*BITS.
                const   EXPTR(long) p1,         // Multiplicand, shift table.
                const   EXPTR(long) p2,         // Unsigned multiplier.
                        size_t      k,          // Left shift of product.
                        EXPTR(long) p           // Augend in buffer of size
                                                // max (*p, *p1 + *p2 + k) + 2.
        )                                       // p1, p2 may overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

//      Algorithm with table of shifts.

  if (n1 >= cBigNumberSkipLow0 (p1))            // Multiplicand is not empty.
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);       // Denormalization.
    n1 += 2;                                    // Step of table of shifts.
#ifndef _CBIGNUM_TERNARY_MUL
    for (size_t k2 = 0; k2 != n2; k2++)         // Cycle on words of p2.
    {
      long num = p2 [k2 + 1];                   // Current word of p2.
      long mask = 1;                            // Mask for bits of num.
      const EXPTR(long) pp1 = p1;               // Current shift.
      for (;; pp1 += n1)                        // Cycle on bits of word of p2.
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k + k2);
        if ((mask <<= 1) == 0) break;           // No more bits.
      }                                         // End of cycle on bits of p2.
    }                                           // End of cycle on words of p2.
#else //_CBIGNUM_TERNARY_MUL
    long num2 = p2 [n2], num1 = num2;           // High bit of the high word.
    num2 <<= 1;                                 // Next bit.
    long mode = 0;                              // Mode 0=add, -1=subtract.
    if ((num1 & num2) < 0)                      // Both bits are 1.
    {
      cBigNumberMAddM (p, p1, k + n2);          // Addition with shift n2.
      num1 = ~num1; num2 = ~num2; mode = ~0;    // Switch to subtract mode.
    }
    p1 += n1 * BITS; p1 -= n1;                  // Initial shift.
    for (size_t k2 = n2; k2-- != 0;)            // Cycle on words of p2.
    {
      const EXPTR(long) pp1 = p1;               // Cycle on words of p2.
      for (size_t kb = BITS;; pp1 -= n1)        // Cycle on bits of word of p2.
      {
        long num = num1;                        // Current bit.
        num1 = num2;                            // 1st next bit.
        if (kb != 2) num2 <<= 1;                // 2nd next bit.
        else num2 = (k2? p2 [k2]: 0L) ^ mode;   // Load from next word.
        if ((num | (num1 & num2) ) < 0)         // Check if to accumulate.
        {
           if (mode == 0) cBigNumberMAddM (p, pp1, k + k2);
           else           cBigNumberMSubM (p, pp1, k + k2);
           if (num >= 0) { num1 = ~num1; num2 = ~num2; mode = ~mode; }
        }                                       // Switch mode.
        if (--kb == 0) break;                   // Mo more bits.
      }                                         // End of cycle on bits of p2.
    }                                           // End of cycle on words of p2.
#endif//_CBIGNUM_TERNARY_MUL
  }

  cBigNumberFit (p);                            // Normalization.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

void    cBigNumberMSubMulShlTab (               // Subtract of multiplication
                                                // p -= p1 * p2 << k*BITS.
                const   EXPTR(long) p1,         // Multiplicand, shift table.
                const   EXPTR(long) p2,         // Unsigned multiplier.
                        size_t      k,          // Left shift of product.
                        EXPTR(long) p           // Subtrahend in buffer of size
                                                // max (*p, *p1 + *p2 + k) + 2.
        )                                       // p1, p2 may overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

//      Algorithm with table of shifts.

  if (n1 >= cBigNumberSkipLow0 (p1))            // Multiplicand is not empty.
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);       // Denormalization.
    n1 += 2;                                    // Step of table of shifts.
#ifndef _CBIGNUM_TERNARY_MUL
    for (size_t k2 = 0; k2 != n2; k2++)         // Cycle on words of p2.
    {
      long num = p2 [k2 + 1];                   // Current word of p2.
      long mask = 1;                            // Mask for bits of num.
      const EXPTR(long) pp1 = p1;               // Current shift.
      for (;; pp1 += n1)                        // Cycle on bits of word of p2.
      {
        if (num & mask) cBigNumberMSubM (p, pp1, k + k2);
        if ((mask <<= 1) == 0) break;           // No more bits.
      }                                         // End of cycle on bits of p2.
    }                                           // End of cycle on words of p2.
#else //_CBIGNUM_TERNARY_MUL
    long num2 = p2 [n2], num1 = num2;           // High bit of the high word.
    num2 <<= 1;                                 // Next bit.
    long mode = 0;                              // Mode 0=subtract, -1=add.
    if ((num1 & num2) < 0)                      // Both bits are 1.
    {
      cBigNumberMSubM (p, p1, k + n2);          // Subtraction with shift n2.
      num1 = ~num1; num2 = ~num2; mode = ~0;    // Switch to add mode.
    }
    p1 += n1 * BITS; p1 -= n1;                  // Initial shift.
    for (size_t k2 = n2; k2-- != 0;)            // Cycle on words of p2.
    {
      const EXPTR(long) pp1 = p1;               // Cycle on words of p2.
      for (size_t kb = BITS;; pp1 -= n1)        // Cycle on bits of word of p2.
      {
        long num = num1;                        // Current bit.
        num1 = num2;                            // 1st next bit.
        if (kb != 2) num2 <<= 1;                // 2nd next bit.
        else num2 = (k2? p2 [k2]: 0L) ^ mode;   // Load from next word.
        if ((num | (num1 & num2) ) < 0)         // Check if to accumulate.
        {
           if (mode == 0) cBigNumberMSubM (p, pp1, k + k2);
           else           cBigNumberMAddM (p, pp1, k + k2);
           if (num >= 0) { num1 = ~num1; num2 = ~num2; mode = ~mode; }
        }                                       // Switch mode.
        if (--kb == 0) break;                   // Mo more bits.
      }                                         // End of cycle on bits of p2.
    }                                           // End of cycle on words of p2.
#endif//_CBIGNUM_TERNARY_MUL
  }

  cBigNumberFit (p);                            // Normalization.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

#ifndef _CBIGNUM_HARDWARE_MUL                   // If not in Cbignumf.cpp.

//      Special functions for accumulation of multiplication with shift.
//      To speed up operation, functions use either Karatsuba method or
//      method of block multiplication and build temporary tables of shifts.
//
//      Multiplier is considered to be UNSIGNED number,
//      i.e. sign bits are interpreted as high bits.
//
//      Functions does not require for normalization of operands
//      and always return normalized result.

void    cBigNumberMAddMulShlKar (               // Addition of multiplication
                                                // p += p1 * p2 << k*BITS.
                        EXPTR(long) p1,         // Multiplicand in overwritten
                                                // buffer of size *p1 + 3 +
                                                // *p2 < _CBNL_TAB_MIN?
                                                //  (min (*p1, _CBNL_TAB_HIGH)
                                                //   + 2) * (BITS - 1).
                        EXPTR(long) p2,         // Unsigned multiplier.
                        size_t      k,          // Left shift of product.
                        EXPTR(long) p           // Augend in buffer of size
                                                // max (*p, *p1 + *p2 + k + 1)
                                                //                        + 2.
        )                                       // p1, p2, p may not overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

//      Check if Karatsuba method is applicable.

#ifdef  _CBIGNUM_KARATSUBA_MUL
  if (n1 > _CBNL_KARATSUBA_MIN && n2 > _CBNL_KARATSUBA_MIN)
  {

//      Multiply large 1st operand by pieces.
//      NOTE: Algorithm requires to keep p2 intact in the recursive call.

    while (n1 > n2 * 2 - n2 / 4)                // Check of size.
    {
      long lp = p1 [n1 -= n2];                  // Save word to be overwritten.
      p1 [n1] = (long) n2;                      // Put size of piece.
      cBigNumberMAddMulShlKar (p1 + n1, p2, k + n1, p);
      p1 [n1] = lp;                             // Restore word.
      p1 [n1 + 1] = 0; n1 += (lp < 0);          // Add high 0-word.
      p1 [0] = (long) n1;                       // Put size of last piece.
    }

//      Multiply operands of similar sizes by Karatsuba.

    size_t n = (n1 >= n2? n1: n2 - 1) / 2;      // Size of low part.

//      NOTE: To prevent overflow of p we must keep high part of p2
//            larger then low part of p2 for rare special case when
//            p1 is twice or more shorter then p2.

    cBigTemp c12; c12.checkexpand (n * 2 + 5);
    EXPTR(long) p12 = EXPTRTYPE(c12);           // Buffer for multiplication.

    cBigTemp c1; c1.checkexpand (n + 2);
    cBigTemp c2; c2.checkexpand (n + 4);
    cBigNumberCopyShr (p2, n, EXPTRTYPE(c2));   // Get high part from p2 to c2.

//      Size of p1 may be n or smaller if p1 is twice or more shorter then p2.
//      NOTE: Algorithm requires to keep c2 intact in the recursive call.

    if (n1 > n)
    {
      cBigNumberCopyShr (p1, n, EXPTRTYPE(c1)); // Get high part from p1 to c1.
      long lp = p1 [n];                         // Save word to be overwritten.
      p1 [n] = (long)(n1 - n);                  // Put size of piece.
      p12 [0] = 0;                              // Clear multiply accumulator.
      cBigNumberMAddMulShlKar (p1 + n, EXPTRTYPE(c2), 0, p12);
                                                // Multiply c1 * c2.
      cBigNumberMAddShl (p, p12, k + n * 2);    // Accumulate c1 * c2.
      cBigNumberMSubShl (p, p12, k + n);        // Accumulate c1 * c2.
      p1 [n] = lp;                              // Restore word.
      p1 [n + 1] = 0;                           // Add high 0-word to p1.
      p1 [0] = (long)(n + (lp < 0));            // Get low part of p1.
      cBigNumberMAddShl (EXPTRTYPE(c1), p1, 0); // Add c1 + p1.
    }
    else
      cBigNumberCopy (p1, EXPTRTYPE(c1));       // Get copy of p1 to c1.

//      Size of p2 can not be n or smaller.

    assert (n2 > n);
    {
      size_t nc2 = c2.length() + 1;             // Add high 0-word to c2.
      c2.item (0) = (long) nc2;
      ((EXPTR(long)) EXPTRTYPE(c2)) [nc2] = 0;

      long lp = p2 [n + 1];                     // Save word to be overwritten.
      p2 [n + 1] = 0;                           // Add high 0-word to p2.
      p2 [0] = (long)(n + 1);                   // Get low part of p2.
      cBigNumberMAddShl (EXPTRTYPE(c2), p2, 0); // Add c2 + p2.
      p2 [n + 1] = lp;                          // Restore word.

      p2 [0] = (long) n;                        // Remove high 0-word from p2.
      p12 [0] = 0;                              // Clear multiply accumulator.
      cBigNumberMAddMulShlKar (p1, p2, 0, p12); // Multiply p1 * p2.
      p2 [0] = (long) n2;                       // Restore original p2.

      c2.item (0) = (long)(c2.length() - (c2.hiword() == 0));
                                                // Remove high 0-word from c2.
      cBigNumberCopy (EXPTRTYPE(c1), p1);       // Accumulate (c1+p2)*(c2+p2).
      cBigNumberMAddMulShlKar (p1, EXPTRTYPE(c2), k + n, p);

      cBigNumberMSubShl (p, p12, k + n);        // Accumulate p1 * p2.
      cBigNumberMAddShl (p, p12, k);            // Accumulate p1 * p2.
    }

    return;
  }
#endif//_CBIGNUM_KARATSUBA_MUL

//      Multiply 1st operand by pieces if 2nd operand is small.
//
//      The high piece may be either positive or negative, but lower
//      pieces are all positive, so we must add high 0-word to them.
//      The high and lower pieces contains _CBNL_TAB_OPT + 1 words,
//      including the high 0-word addendum for lower pieces, except
//      for the lowest piece that may contains up to _CBNL_TAB_HIGH
//      word with the high 0-word addendum.
//
//      Algorithm extracts pieces in place without copying.

#ifdef  _CBIGNUM_BLOCK_MUL
  if (n1 > _CBNL_TAB_HIGH)
  {
    --n1;                                       // Get high word of high piece.
    do                                          // Cycle on pieces.
    {
      n1 -= _CBNL_TAB_OPT;                      // Total size of lower pieces.
      long lp = p1 [n1];                        // Save word to be overwritten.
      p1 [n1] = _CBNL_TAB_OPT + 1;              // and put size of piece.
#ifdef  _CBIGNUM_SHIFTTAB_MUL
      if (n2 >= _CBNL_TAB_MIN) {                // Use table of shifts?
        cBigNumberTab (p1 + n1);                // Prepare table of shifts.
        cBigNumberMAddMulShlTab (p1 + n1, p2, k + n1, p);
      } else
#endif//_CBIGNUM_SHIFTTAB_MUL
        cBigNumberMAddMulShl (p1 + n1, p2, k + n1, p);
      p1 [n1 + 1] = 0;                          // Add high 0-word to piece.
      p1 [n1] = lp;                             // Restore word.
    }
    while (n1 >= _CBNL_TAB_HIGH);
    p1 [0] = (long)(n1 + 1);                    // Put size of lowest piece.
  }
#endif//_CBIGNUM_BLOCK_MUL

//      Multiply the lowest piece.

#ifdef  _CBIGNUM_SHIFTTAB_MUL
#ifdef  _CBIGNUM_BLOCK_MUL
  if (n2 >= _CBNL_TAB_MIN) {                    // Use table of shifts?
#else
  if (n2 >= _CBNL_TAB_MIN && n1 <= _CBNL_TAB_HIGH) {
#endif//_CBIGNUM_BLOCK_MUL
    cBigNumberTab (p1);                         // Prepare table of shifts.
    cBigNumberMAddMulShlTab (p1, p2, k, p);
  } else
#endif//_CBIGNUM_SHIFTTAB_MUL
    cBigNumberMAddMulShl (p1, p2, k, p);
}

void    cBigNumberMSubMulShlKar (               // Subtract of multiplication
                                                // p -= p1 * p2 << k*BITS.
                        EXPTR(long) p1,         // Multiplicand in overwritten
                                                // buffer of size *p1 + 3 +
                                                // *p2 < _CBNL_TAB_MIN?
                                                //  (min (*p1, _CBNL_TAB_HIGH)
                                                //   + 2) * (BITS - 1).
                        EXPTR(long) p2,         // Unsigned multiplier.
                        size_t      k,          // Left shift of product.
                        EXPTR(long) p           // Subtrahend in buffer of size
                                                // max (*p, *p1 + *p2 + k + 1)
                                                //                        + 2.
        )                                       // p1, p2, p may not overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

//      Check if Karatsuba method is applicable.

#ifdef  _CBIGNUM_KARATSUBA_MUL
  if (n1 > _CBNL_KARATSUBA_MIN && n2 > _CBNL_KARATSUBA_MIN)
  {

//      Multiply large 1st operand by pieces.
//      NOTE: Algorithm requires to keep p2 intact in the recursive call.

    while (n1 > n2 * 2 - n2 / 4)                // Check of size.
    {
      long lp = p1 [n1 -= n2];                  // Save word to be overwritten.
      p1 [n1] = (long) n2;                      // Put size of piece.
      cBigNumberMSubMulShlKar (p1 + n1, p2, k + n1, p);
      p1 [n1] = lp;                             // Restore word.
      p1 [n1 + 1] = 0; n1 += (lp < 0);          // Add high 0-word.
      p1 [0] = (long) n1;                       // Put size of last piece.
    }

//      Multiply operands of similar sizes by Karatsuba.

    size_t n = (n1 >= n2? n1: n2 - 1) / 2;      // Size of low part.

//      NOTE: To prevent overflow of p we must keep high part of p2
//            larger then low part of p2 for rare special case when
//            p1 is twice or more shorter then p2.

    cBigTemp c12; c12.checkexpand (n * 2 + 5);
    EXPTR(long) p12 = EXPTRTYPE(c12);           // Buffer for multiplication.

    cBigTemp c1; c1.checkexpand (n + 2);
    cBigTemp c2; c2.checkexpand (n + 4);
    cBigNumberCopyShr (p2, n, EXPTRTYPE(c2));   // Get high part from p2 to c2.

//      Size of p1 may be n or smaller if p1 is twice or more shorter then p2.
//      NOTE: Algorithm requires to keep c2 intact in the recursive call.

    if (n1 > n)
    {
      cBigNumberCopyShr (p1, n, EXPTRTYPE(c1)); // Get high part from p1 to c1.
      long lp = p1 [n];                         // Save word to be overwritten.
      p1 [n] = (long)(n1 - n);                  // Put size of piece.
      p12 [0] = 0;                              // Clear multiply accumulator.
      cBigNumberMAddMulShlKar (p1 + n, EXPTRTYPE(c2), 0, p12);
                                                // Multiply c1 * c2.
      cBigNumberMSubShl (p, p12, k + n * 2);    // Accumulate c1 * c2.
      cBigNumberMAddShl (p, p12, k + n);        // Accumulate c1 * c2.
      p1 [n] = lp;                              // Restore word.
      p1 [n + 1] = 0;                           // Add high 0-word to p1.
      p1 [0] = (long)(n + (lp < 0));            // Get low part of p1.
      cBigNumberMAddShl (EXPTRTYPE(c1), p1, 0); // Add c1 + p1.
    }
    else
      cBigNumberCopy (p1, EXPTRTYPE(c1));       // Get copy of p1 to c1.

//      Size of p2 can not be n or smaller.

    assert (n2 > n);
    {
      size_t nc2 = c2.length() + 1;             // Add high 0-word to c2.
      c2.item (0) = (long) nc2;
      ((EXPTR(long)) EXPTRTYPE(c2)) [nc2] = 0;

      long lp = p2 [n + 1];                     // Save word to be overwritten.
      p2 [n + 1] = 0;                           // Add high 0-word to p2.
      p2 [0] = (long)(n + 1);                   // Get low part of p2.
      cBigNumberMAddShl (EXPTRTYPE(c2), p2, 0); // Add c2 + p2.
      p2 [n + 1] = lp;                          // Restore word.

      p2 [0] = (long) n;                        // Remove high 0-word from p2.
      p12 [0] = 0;                              // Clear multiply accumulator.
      cBigNumberMAddMulShlKar (p1, p2, 0, p12); // Multiply p1 * p2.
      p2 [0] = (long) n2;                       // Restore original p2.

      c2.item (0) = (long)(c2.length() - (c2.hiword() == 0));
                                                // Remove high 0-word from c2.
      cBigNumberCopy (EXPTRTYPE(c1), p1);       // Accumulate (c1+p2)*(c2+p2).
      cBigNumberMSubMulShlKar (p1, EXPTRTYPE(c2), k + n, p);

      cBigNumberMAddShl (p, p12, k + n);        // Accumulate p1 * p2.
      cBigNumberMSubShl (p, p12, k);            // Accumulate p1 * p2.
    }

    return;
  }
#endif//_CBIGNUM_KARATSUBA_MUL

//      Multiply 1st operand by pieces if 2nd operand is small.
//
//      The high piece may be either positive or negative, but lower
//      pieces are all positive, so we must add high 0-word to them.
//      The high and lower pieces contains _CBNL_TAB_OPT + 1 words,
//      including the high 0-word addendum for lower pieces, except
//      for the lowest piece that may contains up to _CBNL_TAB_HIGH
//      word with the high 0-word addendum.
//
//      Algorithm extracts pieces in place without copying.

#ifdef  _CBIGNUM_BLOCK_MUL
  if (n1 > _CBNL_TAB_HIGH)
  {
    --n1;                                       // Get high word of high piece.
    do                                          // Cycle on pieces.
    {
      n1 -= _CBNL_TAB_OPT;                      // Total size of lower pieces.
      long lp = p1 [n1];                        // Save word to be overwritten.
      p1 [n1] = _CBNL_TAB_OPT + 1;              // and put size of piece.
#ifdef  _CBIGNUM_SHIFTTAB_MUL
      if (n2 >= _CBNL_TAB_MIN) {                // Use table of shifts?
        cBigNumberTab (p1 + n1);                // Prepare table of shifts.
        cBigNumberMSubMulShlTab (p1 + n1, p2, k + n1, p);
      } else
#endif//_CBIGNUM_SHIFTTAB_MUL
        cBigNumberMSubMulShl (p1 + n1, p2, k + n1, p);
      p1 [n1 + 1] = 0;                          // Add high 0-word to piece.
      p1 [n1] = lp;                             // Restore word.
    }
    while (n1 >= _CBNL_TAB_HIGH);
    p1 [0] = (long)(n1 + 1);                    // Put size of lowest piece.
  }
#endif//_CBIGNUM_BLOCK_MUL

//      Multiply the lowest piece.

#ifdef  _CBIGNUM_SHIFTTAB_MUL
#ifdef  _CBIGNUM_BLOCK_MUL
  if (n2 >= _CBNL_TAB_MIN) {                    // Use table of shifts?
#else
  if (n2 >= _CBNL_TAB_MIN && n1 <= _CBNL_TAB_HIGH) {
#endif//_CBIGNUM_BLOCK_MUL
    cBigNumberTab (p1);                         // Prepare table of shifts.
    cBigNumberMSubMulShlTab (p1, p2, k, p);
  } else
#endif//_CBIGNUM_SHIFTTAB_MUL
    cBigNumberMSubMulShl (p1, p2, k, p);
}

//      Universal functions of multiplication work for numbers
//      of any signs. To speed up operation, functions may use
//      Karatsuba method and build temporary tables of shifts.
//
//      For better performance considerably larger operand should
//      be placed first.
//
//      Functions does not require for normalization of operands
//      and always return normalized result.

void    cBigNumberMAddMul (                     // Addition of multiplication
                                                // p += p1 * p2.
                const   EXPTR(long) p1,         // Multiplicand.
                const   EXPTR(long) p2,         // Multiplier.
                        EXPTR(long) p           // Augend in buffer of size
                                                // max (*p, *p1 + *p2 + 1) + 2.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

  cBigTemp cBigBuf1;                            // Allocate temporary buffer.
  cBigTemp cBigBuf2;                            // Allocate temporary buffer.
  cBigBuf1.checkexpand ((n2 < _CBNL_TAB_MIN)? (n1 + 2):
                        (n1 < _CBNL_TAB_HIGH)? ((n1 + 2) * BITS + 1):
                          (n1 + 2 + (_CBNL_TAB_HIGH + 2) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);                // Memory allocation.
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);        // Buffer for multiplicand.
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);        // Buffer for multiplier.

  cBigNumberCopy (p1, pp1);                     // Copying.

  if (p2 [n2] >= 0)                             // Not negative multiplier:
  {
    cBigNumberCopy (p2, pp2);                   // Copying.
    cBigNumberMAddMulShlKar (pp1, pp2, 0, p);   // Optimized multiplication.
  }
  else                                          // Negative multiplier:
  {
    cBigNumberNeg (p2, pp2);                    // Sign inversion.
    *pp2 = (long)n2;                            // Do not increase size.
    cBigNumberMSubMulShlKar (pp1, pp2, 0, p);   // Optimized multiplication.
  }
}

void    cBigNumberMSubMul (                     // Subtract of multiplication
                                                // p -= p1 * p2.
                const   EXPTR(long) p1,         // Multiplicand.
                const   EXPTR(long) p2,         // Multiplier.
                        EXPTR(long) p           // Subtrahend in buffer of size
                                                // max (*p, *p1 + *p2 + 1) + 2.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

  cBigTemp cBigBuf1;                            // Allocate temporary buffer.
  cBigTemp cBigBuf2;                            // Allocate temporary buffer.
  cBigBuf1.checkexpand ((n2 < _CBNL_TAB_MIN)? (n1 + 2):
                        (n1 < _CBNL_TAB_HIGH)? ((n1 + 2) * BITS + 1):
                          (n1 + 2 + (_CBNL_TAB_HIGH + 2) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);                // Memory allocation.
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);        // Buffer for multiplicand.
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);        // Buffer for multiplier.

  cBigNumberCopy (p1, pp1);                     // Copying.

  if (p2 [n2] >= 0)                             // Not negative multiplier:
  {
    cBigNumberCopy (p2, pp2);                   // Copying.
    cBigNumberMSubMulShlKar (pp1, pp2, 0, p);   // Optimized multiplication.
  }
  else                                          // Negative multiplier:
  {
    cBigNumberNeg (p2, pp2);                    // Sign inversion.
    *pp2 = (long)n2;                            // Do not increase size.
    cBigNumberMAddMulShlKar (pp1, pp2, 0, p);   // Optimized multiplication.
  }
}

void    cBigNumberMul (                         // Multiplication p = p1 * p2.
                const   EXPTR(long) p1,         // Multiplicand.
                const   EXPTR(long) p2,         // Multiplier.
                        EXPTR(long) p           // Buffer of size
                                                // *p1 + *p2 + 3.
        )                                       // p1, p2, p may overlap.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.

  cBigTemp cBigBuf1;                            // Allocate temporary buffer.
  cBigTemp cBigBuf2;                            // Allocate temporary buffer.
  cBigBuf1.checkexpand ((n2 < _CBNL_TAB_MIN)? (n1 + 2):
                        (n1 < _CBNL_TAB_HIGH)? ((n1 + 2) * BITS + 1):
                          (n1 + 2 + (_CBNL_TAB_HIGH + 2) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);                // Memory allocation.
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);        // Buffer for multiplicand.
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);        // Buffer for multiplier.

  cBigNumberCopy (p1, pp1);                     // Copying.

//      Clearing of result is the only difference between cBigNumberMul()
//      and cBigNumberMAddMul(). We must clear the result after copying
//      of numbers because operands may overlap.

  if (p2 [n2] >= 0)                             // Not negative multiplier:
  {
    cBigNumberCopy (p2, pp2);                   // Copying.
    p [0] = 0;                                  // Clearing of result.
    cBigNumberMAddMulShlKar (pp1, pp2, 0, p);   // Optimized multiplication.
  }
  else                                          // Negative multiplier:
  {
    cBigNumberNeg (p2, pp2);                    // Sign inversion.
    *pp2 = (long)n2;                            // Do not increase size.
    p [0] = 0;                                  // Clearing of result.
    cBigNumberMSubMulShlKar (pp1, pp2, 0, p);   // Optimized multiplication.
  }
}

//      Function of long multiplication.

void    cLongMul        (                       // Multiplication l1 * l2.
                        long    l1,             // Multiplicand.
                        long    l2,             // Multiplier.
                        long    p [2]           // Result.
        )
{
  long c1 [5]; c1 [0] = 1; c1 [1] = l1;         // Buffer for multiplicand.
  long c2 [3]; c2 [0] = 1; c2 [1] = l2;         // Buffer for multiplier.
  long c  [5]; c [0] = c [1] = c [2] = 0;       // Buffer for result.
  EXPTR(long) pp1 = EXPTRTO(long,c1,sizeof(c1)/sizeof(*c1)-1);
  EXPTR(long) pp2 = EXPTRTO(long,c2,sizeof(c2)/sizeof(*c2)-1);
  EXPTR(long) pp  = EXPTRTO(long,c, sizeof(c) /sizeof(*c) -1);

  if (l2 >= 0)                                  // Not negative multiplier:
    cBigNumberMAddMulShl (pp1, pp2, 0, pp);     // Multiplication.
  else {                                        // Negative multiplier:
    c2 [1] = -l2;                               // Sign inversion.
    cBigNumberMSubMulShl (pp1, pp2, 0, pp);     // Multiplication.
  }

  p [0] = pp [1];
  p [1] = pp [2];
}

//      Function of unsigned long multiplication.

void    cULongMul       (                       // Multiplication l1 * l2.
               unsigned long    l1,             // Unsigned multiplicand.
               unsigned long    l2,             // Unsigned multiplier.
               unsigned long    p [2]           // Unsigned result.
        )
{                                               // Buffer for multiplicand.
  long c1 [5]; c1 [0] = 2; c1 [1] = l1; c1 [2] = 0;
  long c2 [3]; c2 [0] = 1; c2 [1] = l2;         // Buffer for multiplier.
  long c  [6]; c [0] = c [1] = c [2] = 0;       // Buffer for result.
  EXPTR(long) pp1 = EXPTRTO(long,c1,sizeof(c1)/sizeof(*c1)-1);
  EXPTR(long) pp2 = EXPTRTO(long,c2,sizeof(c2)/sizeof(*c2)-1);
  EXPTR(long) pp  = EXPTRTO(long,c, sizeof(c) /sizeof(*c) -1);

  cBigNumberMAddMulShl (pp1, pp2, 0, pp);       // Multiplication.

  p [0] = pp [1];
  p [1] = pp [2];
}

#endif//_CBIGNUM_HARDWARE_MUL

//================================================
//      Function handling division by 0.
//================================================

size_t  cBigNumberMaskDiv0;                     // Mask for division by 0:
                                                //  0 - Divide Error
void    cBigNumberDiv0()                        // >0 - Ignore
{
  cBigNumberMaskDiv0 /= cBigNumberMaskDiv0;
}

//================================================
//      Functions of integral division.
//================================================

//      Special function of division with table of shifts.
//      It is assumed, that dividend and divider have the same signs
//      except for case when dividend is negative and divider is 0.
//      Buffer of divider must contain table of shifts,
//      obtained by function cBigNumberTab.
//
//      Function gets quotient and assigns module in place of dividend.
//      Quotient >= 0. Sign of module is same as sign of dividend.
//      On division by 0, if allowed, quotient is 0 and module is
//      same as dividend.
//
//      Dividend must be normalized.
//      Function returns normalized quotient and module.

void    cBigNumberMModDivTab (                  // Division p = p1 / p2,
                                                //          p1 %= p2.
                        EXPTR(long) p1,         // Dividend, then module.
                const   EXPTR(long) p2,         // Divider and table of shifts.
                        EXPTR(long) p           // Buffer of size *p1 + 2.
        )                                       // p1, p2, p may not overlap.
{
  assert (cBigNumberIsFit (p1));                // Check of normalization.
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  assert (p2 != p1);                            // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;       // Word counter in quotient.

//      We must add 0-word to quotient for case when the high bit
//      of quotient is set to 1. If 0-word is redundant it will be
//      deleted on normalization.

  *p = (long)(n + 1); p += n; p [1] = 0;        // Start filling of p.

//      Build next words.

  if (n2 >= cBigNumberSkipLow0 (p2))            // Skip low 0-words.
  {                                             // Check for division by 0.
    assert ((p1 [n1] ^ p2 [n2]) >= 0);          // Check if signs are the same.
    long lt = (p1 [n1] < 0) * 2 - 1;            // Constant to compare
                                                // absolute values of numbers.
                                                // (p1 and p2 >= 0)? -1: 1
    n2 += 2;                                    // Step of table of shifts.
    do                                          // Cycle on words of quotient.
    {
      --n;                                      // Word counter in quotient.
      p2 += n2 * BITS;                          // Word shift of divider.
      unsigned long mask = (unsigned long)
                           (1L << (BITS-1));    // Initial mask.
      *p = 0;                                   // Clearing word of quotient.
      do                                        // Cycle on bits of quotient.
      {
        p2 -= n2;                               // Current shift.
        long diff = (long)n - *p1 + *p2;        // Difference of sizes.
        if (diff <= 0 && (diff < 0 ||           // If remainder not less
            cBigNumberCompHigh (p1, p2) != lt)) // then shifted divider,
        {                                       // then
          cBigNumberMSubD (p1, p2, n);          // subtract divider and
          (*p) += mask;                         // set bit of quotient.
        }
      }
      while ((mask >>= 1) != 0);                // End of cycle on bits.
      --p;                                      // Word of quotient is ready.
    }
    while (n);                                  // End of cycle on words.
    assert ((p1 [(size_t)(*p1)] ^ lt) < 0 ||    // Check if signs are the same
            (p1 [(size_t)(*p1)] + *p1) == 1);   // or remainder is 0.
    assert (cBigNumberComp (p1, p2) == lt);     // Check if remainder<divider.
    assert (cBigNumberIsFit (p1));              // Check of normalization.
  }
  else                                          // Divider is 0.
  {                                             // If division by 0
    cBigNumberDiv0();                           // is enabled then module
    do { *p-- = 0; } while (--n);               // is same as dividend.
  }

  cBigNumberFit (p);                            // Normalization of quotient.
  assert (p [(size_t)(*p)] >= 0);               // Check if non-negative.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

//      Universal function of division with module works for
//      numbers of any signs and does not require for preliminary
//      prepared table of shifts. If it may speed up the operation,
//      the function builds temporary table of shifts, otherwise
//      it applies algorithm, which does not use table of shifts.
//
//      Function gets quotient and assigns module in place of dividend.
//      Sign of quotient is production of signs of dividend and divider.
//      Sign of module is same as sign of dividend.
//      On division by 0, if allowed, quotient is 0 and module is
//      same as dividend.
//
//      Dividend must be normalized.
//      Function returns normalized quotient and module.

void    cBigNumberMModDiv (                     // Division p = p1 / p2,
                                                //          p1 %= p2.
                        EXPTR(long) p1,         // Dividend, then module.
                const   EXPTR(long) p2,         // Divider.
                        EXPTR(long) p           // Buffer of size *p1 + 3.
        )                                       // p2 may overlap with p1, p.
{
  assert (p1 != p);                             // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.
  long nt = (long)((n2 - n1 + _CBNL_TAB_MIN - 1) & (n2 - _CBNL_TAB_MAX));

  cBigTemp cBigBuf2;                            // Allocate temporary buffer.
  cBigBuf2.checkexpand ((size_t)((nt < 0)? ((n2 + 3) * BITS + 1): (n2 + 3)));
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);        // Buffer for divider.

//      Providing for same signs in dividend and divider.

  long signdiv;                                 // Sign of non-0 quotient.
  if ((p1 [n1] ^ p2 [n2]) >= 0)                 // Signs are the same:
    { cBigNumberCopy (p2, pp2); signdiv = 1; }  // copying of divider.
  else                                          // Signs are not the same:
    { cBigNumberNeg (p2, pp2); signdiv = -1; }  // sign inversion.

//      Algorithm with table of shifts.

#ifdef  _CBIGNUM_SHIFTTAB_DIV
  if (nt < 0)
  {
    cBigNumberTab (pp2);                        // Prepare table of shifts.
    cBigNumberMModDivTab (p1, pp2, p);          // Division.
    if (signdiv < 0) cBigNumberNeg (p, p);      // Sign inversion.
    return;
  }
#endif//_CBIGNUM_SHIFTTAB_DIV

//      Algorithm that does not require for table of shifts.

  assert (cBigNumberIsFit (p1));                // Check of normalization.
  n2 = cBigNumberFit (pp2);                     // Normalization of divider.
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;       // Word counter in quotient.

//      We must add 0-word to quotient for case when the high bit
//      of quotient is set to 1. If 0-word is redundant it will be
//      deleted on normalization.

  *p = (long)(n + 1); p += n; p [1] = 0;        // Start filling of p.

//      Build next words.

  if (n2 >= cBigNumberSkipLow0 (p2))            // Skip low 0-words.
  {                                             // Check for division by 0.
    assert ((p1 [n1] ^ pp2 [n2]) >= 0);         // Check if signs are the same.
    long lt = (p1 [n1] < 0) * 2 - 1;            // Constant to compare
                                                // absolute values of numbers
                                                // (p1 and pp2 >= 0)? -1: 1
    do                                          // Cycle on words of quotient.
    {
      --n;                                      // Word counter in quotient.
      *p = 0;                                   // Clearing word of quotient.
      cBigNumberMShlD (pp2);                    // Word shift of divider.
      unsigned long mask = (1UL << (BITS-1));   // Initial mask.
      do                                        // Cycle on bits of quotient.
      {
        long diff = (long)cBigNumberMDiv2D (pp2)// Bit shift of divider.
                  + (long)n - *p1;              // Difference of sizes.
        if (diff <= 0 && (diff < 0 ||           // If remainder not less
            cBigNumberCompHigh (p1, pp2) != lt))// then shifted divider,
        {                                       // then
          cBigNumberMSubD (p1, pp2, n);         // subtract divider and
          (*p) += mask;                         // set bit of quotient.
        }
      }
      while ((mask >>= 1) != 0);                // End of cycle on bits.
      --p;                                      // Word of quotient is ready.
    }
    while (n);                                  // End of cycle on words.
    assert ((p1 [(size_t)(*p1)] ^ lt) < 0 ||    // Check if signs are the same
            (p1 [(size_t)(*p1)] + *p1) == 1);   // or remainder is 0.
    assert (cBigNumberComp (p1, pp2) == lt);    // Check if remainder<divider.
    assert (cBigNumberIsFit (p1));              // Check of normalization.
  }
  else                                          // Divider is 0.
  {                                             // If division by 0
    cBigNumberDiv0();                           // is enabled then module
    do { *p-- = 0; } while (--n);               // is same as dividend.
  }

  cBigNumberFit (p);                            // Normalization of quotient.
  assert (p [(size_t)(*p)] >= 0);               // Check if non-negative.
  if (signdiv < 0) cBigNumberNeg (p, p);        // Sign inversion.

#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

//================================================
//      Functions of module.
//================================================

//      Special function of module with table of shifts.
//      It is assumed, that dividend and divider have the same signs
//      except for case when dividend is negative and divider is 0.
//      Buffer of divider must contain table of shifts,
//      obtained by function cBigNumberTab.
//
//      Function of module is reduced variant of function of
//      division after exclusion of calculation of quotient.
//      Function assigns module in place of dividend.
//      Sign of module is same as sign of dividend.
//      On division by 0, if allowed, module is same as dividend.
//
//      Dividend must be normalized.
//      Function returns normalized module.

void    cBigNumberMModTab (                     // Module p1 %= p2.
                        EXPTR(long) p1,         // Dividend, then module.
                const   EXPTR(long) p2          // Divider and table of shifts.
        )                                       // p1, p2 may not overlap.
{
  assert (cBigNumberIsFit (p1));                // Check of normalization.
  assert (p1 != p2);                            // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;       // Word counter in quotient.

  if (n2 >= cBigNumberSkipLow0 (p2))            // Skip low 0-words.
  {                                             // Check for division by 0.
    assert ((p1 [n1] ^ p2 [n2]) >= 0);          // Check if signs are the same.
    long lt = (p1 [n1] < 0) * 2 - 1;            // Constant to compare
                                                // absolute values of numbers
                                                // (p1 and p2 >= 0)? -1: 1
    n2 += 2;                                    // Step of table of shifts.
    do                                          // Cycle on words of quotient.
    {
      --n;                                      // Word counter in quotient.
      p2 += n2 * BITS;                          // Word shift of divider.
      unsigned long mask = (1UL << (BITS-1));   // Initial mask.
      do                                        // Cycle on bits of quotient.
      {
        p2 -= n2;                               // Current shift.
        long diff = (long)n - *p1 + *p2;        // Difference of sizes.
        if (diff <= 0 && (diff < 0 ||           // If remainder not less
            cBigNumberCompHigh (p1, p2) != lt)) // then shifted divider,
        {                                       // then
          cBigNumberMSubD (p1, p2, n);          // subtract divider.
        }
      }
      while ((mask >>= 1) != 0);                // End of cycle on bits.
    }
    while (n);                                  // End of cycle on words.
    assert ((p1 [(size_t)(*p1)] ^ lt) < 0 ||    // Check if signs are the same
            (p1 [(size_t)(*p1)] + *p1) == 1);   // or remainder is 0.
    assert (cBigNumberComp (p1, p2) == lt);     // Check if remainder<divider.
    assert (cBigNumberIsFit (p1));              // Check of normalization.
  }
  else                                          // Divider is 0.
  {                                             // If division by 0
    cBigNumberDiv0();                           // is enabled then module
  }                                             // is same as dividend.

#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

//      Universal function of module works for numbers
//      of any signs and does not require for preliminary
//      prepared table of shifts. If it may speed up the operation,
//      the function builds temporary table of shifts, otherwise
//      it applies algorithm, which does not use table of shifts.
//
//      Function of module is reduced variant of function of
//      division after exclusion of calculation of quotient.
//      Function assigns module in place of dividend.
//      Sign of module is same as sign of dividend.
//      On division by 0, if allowed, module is same as dividend.
//
//      Dividend must be normalized.
//      Function returns normalized module.

void    cBigNumberMMod (                        // Module p1 %= p2.
                        EXPTR(long) p1,         // Dividend, then module.
                const   EXPTR(long) p2          // Divider.
        )                                       // p1, p2 may overlap.
{
  size_t n1 = (size_t)(*p1);                    // Number of words.
  size_t n2 = (size_t)(*p2);                    // Number of words.
  long nt = (long)((n2 - n1 + _CBNL_TAB_MIN - 1) & (n2 - _CBNL_TAB_MAX));

  cBigTemp cBigBuf2;                            // Allocate temporary buffer.
  cBigBuf2.checkexpand ((size_t)((nt < 0)? ((n2 + 3) * BITS + 1): (n2 + 3)));
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);        // Buffer for divider.

//      Providing for same signs in dividend and divider.

  if ((p1 [n1] ^ p2 [n2]) >= 0)                 // Signs are the same:
    cBigNumberCopy (p2, pp2);                   // copying of divider.
  else                                          // Signs are not the same:
    cBigNumberNeg  (p2, pp2);                   // sign inversion.

//      Algorithm with table of shifts.

#ifdef  _CBIGNUM_SHIFTTAB_DIV
  if (nt < 0)
  {
    cBigNumberTab (pp2);                        // Prepare table of shifts.
    cBigNumberMModTab (p1, pp2);                // Module.
    return;
  }
#endif//_CBIGNUM_SHIFTTAB_DIV

//      Algorithm that does not require for table of shifts.

  assert (cBigNumberIsFit (p1));                // Check of normalization.
  n2 = cBigNumberFit (pp2);                     // Normalization of divider.
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;       // Word counter in quotient.

  if (n2 >= cBigNumberSkipLow0 (p2))            // Skip low 0-words.
  {                                             // Check for division by 0.
    assert ((p1 [n1] ^ pp2 [n2]) >= 0);         // Check if signs are the same.
    long lt = (p1 [n1] < 0) * 2 - 1;            // Constant to compare
                                                // absolute values of numbers.
                                                // (p1 and pp2 >= 0)? -1: 1
    do                                          // Cycle on words of quotient.
    {
      --n;                                      // Word counter in quotient.
      cBigNumberMShlD (pp2);                    // Word shift of divider.
      unsigned long mask = (1UL << (BITS-1));   // Initial mask.
      do                                        // Cycle on bits of quotient.
      {
        long diff = (long)cBigNumberMDiv2D (pp2)// Bit shift of divider.
                  + (long)n - *p1;              // Difference of sizes.
        if (diff <= 0 && (diff < 0 ||           // If remainder not less
            cBigNumberCompHigh (p1, pp2) != lt))// then shifted divider,
        {                                       // then
          cBigNumberMSubD (p1, pp2, n);         // subtract divider.
        }
      }
      while ((mask >>= 1) != 0);                // End of cycle on bits.
    }
    while (n);                                  // End of cycle on words.
    assert ((p1 [(size_t)(*p1)] ^ lt) < 0 ||    // Check if signs are the same
            (p1 [(size_t)(*p1)] + *p1) == 1);   // or remainder is 0.
    assert (cBigNumberComp (p1, pp2) == lt);    // Check if remainder<divider.
    assert (cBigNumberIsFit (p1));              // Check of normalization.
  }
  else                                          // Divider is 0.
  {                                             // If division by 0
    cBigNumberDiv0();                           // is enabled then module
  }                                             // is same as dividend.

#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT
}

//================================================
//      Functions of integral power.
//================================================

//      Function of power uses buffer for base as work buffer
//      and requires it to be of same size as buffer for result.
//      Both buffers must contain enough space for the result
//      and also 2 auxiliary words for multiplication purposes.
//      Note that if p2 contains more then 1 word and p1 is not one
//      of -1,0,1 then size of buffers will be greater 512 Mbytes.
//
//      Function does not require for normalization of operands
//      and always returns normalized result.

void    cBigNumberPow (                         // p1 in power p2.
                        EXPTR(long) p1,         // Base in overwritten buffer
                                                // of size >= *p + 3.
                const   EXPTR(long) p2,         // Degree.
                        EXPTR(long) p           // Buffer for result
                                                // of size *p + 3.
        )                                       // p1, p2, p may not overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  assert (p2 != p1);                            // Check if not overlap.
  size_t n2 = (size_t)(*p2);                    // Number of words.

  p [0] = 1; p [1] = 1;                         // Initial power is 1.

  if (p2 [n2] < 0)                              // Check if degree is negative.
  {
    cBigNumberMModDiv (p, p1, p1);              // Return integer part 1/p1 or
    if (p1 [1] + (p2 [1] & 1) < 0) p1 [1] = 1;  // abs(1/p1) if degree is even.
    p [1] = p1 [1];
  }
  else if (n2 != 0)                             // Check if power may be not 1.
  {
    p2++;                                       // Skip to number.
    for (;;)                                    // Cycle on words of p2.
    {
      long num = *p2++;                         // Current word of p2.
      long mask = 1;                            // Mask for bits of num.
      if (--n2 == 0)                            // Reduce number of passes
      {                                         // at the last cycle.
        if (num == 0) break;                    // The high word p2 is 0.
        do                                      // The high bit is always 0.
          mask <<= 1;                           // Shift mask and num, while
        while ((num <<= 1) > 0);                // the high bit of num is 0.
      }
      for (;;)                                  // Cycle on bits of word of p2.
      {
        if (num & mask)
          cBigNumberMul (p1, p, p);             // Accumulation of power.
        if ((mask <<= 1) == 0) break;           // No more bits.
        cBigNumberMul (p1, p1, p1);             // Square of base.
      }                                         // End of cycle on bits of p2.
      if (n2 == 0) break;                       // No more words.
      cBigNumberMul (p1, p1, p1);               // Square of base.
    }                                           // End of cycle on words of p2.
  }
}

//      Function of power by module uses buffer for base as work buffer
//      and requires it to be not smaller then buffer for result.
//      Both buffers must be twice as large as buffer for module
//      plus auxiliary word for multiplication purposes.
//
//      Module must not be 0 (if module is 0 and division by 0 is
//      allowed then result = 1).
//
//      Function does not require for normalization of operands
//      and always returns normalized result.

void    cBigNumberPowMod (                      // p1 in power p2 by mod m
                        EXPTR(long) p1,         // Base in overwritten buffer
                                                // of size >= *mod * 2 + 3.
                const   EXPTR(long) p2,         // Degree.
                const   EXPTR(long) mod,        // Module (may overlap).
                        EXPTR(long) p           // Buffer for result
                                                // of size *mod * 2 + 3.
        )                                       // p1, p2, p may not overlap.
{
  assert (p1 != p);                             // Check if not overlap.
  assert (p2 != p);                             // Check if not overlap.
  assert (p2 != p1);                            // Check if not overlap.
  size_t n2 = (size_t)(*p2);                    // Number of words.

  p [0] = 1; p [1] = 1;                         // Initial power is 1.

  if (p2 [n2] < 0)                              // Check if degree is negative.
  {
    cBigNumberMModDiv (p, p1, p1);              // Power is integral 1/p1 or
    if (p1 [1] + (p2 [1] & 1) < 0) p1 [1] = 1;  // abs(1/p1) if degree is even.
    p [1] = p1 [1]; cBigNumberMMod (p, mod);    // Module of power.
    return;
  }

  size_t nmod = (size_t)(*mod);                 // Number of words.
  cBigTemp cBigBuf3;                            // Allocate temporary buffer.
  cBigBuf3.checkexpand (exmuladd (BITS, nmod, BITS * 3 + 1));
  EXPTR(long) pmod = EXPTRTYPE(cBigBuf3);       // Buffer for module.

//      Preparing of module for division with table of shifts.

  if (mod [nmod] >= 0)
    { cBigNumberCopy (mod, pmod); }             // Copying of module.
  else
    { cBigNumberNeg  (mod, pmod); }             // Sign inversion of module.

  cBigNumberTab (pmod);                         // Prepare table of shifts.
  if (cBigNumberNot0 (pmod) == 0) n2 = 0;       // Division by 0.

  cBigNumberMModTab (p, pmod);                  // Module of power.
  if (n2 != 0 && p [1] != 0)                    // Check if power may change.
  {
    p2++;                                       // Skip to number.

//      Make sign of base positive (for division with table of shifts).

    long sign = p1 [(size_t)(*p1)];             // Sign of base in high bit.
    if (sign < 0)
    {
      cBigNumberNeg (p1, p1);                   // Sign inversion of base.
      sign = (*p2 << (BITS-1));                 // Sign of result in high bit.
    }
    cBigNumberMModTab (p1, pmod);               // Module of base.

//      Power by module.

    for (;;)                                    // Cycle on words of p2.
    {
      long num = *p2++;                         // Current word of p2.
      long mask = 1;                            // Mask for bits of num.
      if (--n2 == 0)                            // Reduce number of passes
      {                                         // at the last cycle.
        if (num == 0) break;                    // The high word p2 is 0.
        do                                      // The high bit is always 0.
          mask <<= 1;                           // Shift mask and num, while
        while ((num <<= 1) > 0);                // the high bit of num is 0.
      }
      for (;;)                                  // Cycle on bits of word of p2.
      {
        if (num & mask)
        {                                       // Accumulation of power.
          if (*p1 > *p) cBigNumberMul (p1, p, p);
          else          cBigNumberMul (p, p1, p);
          cBigNumberMModTab (p, pmod);          // Module of power.
        }
        if ((mask <<= 1) == 0) break;           // No more bits.
        cBigNumberMul (p1, p1, p1);             // Square of base.
        cBigNumberMModTab (p1, pmod);           // Module of square of base.
      }                                         // End of cycle on bits of p2.
      if (n2 == 0) break;                       // No more words.
      cBigNumberMul (p1, p1, p1);               // Square of base.
      cBigNumberMModTab (p1, pmod);             // Module of square of base.
    }                                           // End of cycle on words of p2.
    if (sign < 0) cBigNumberNeg (p, p);         // Sign inversion.
  }
}

//================================================
//      Function of integral square root.
//================================================

//      Function calculated integral square root and assigns
//      remainder in place of base.
//      If number of bits > LONG_MAX cBigNumberERange() is called.
//
//      Base must be normalized.
//      Function always returns normalized root and remainder.

void    cBigNumberMRmSqrt (                     // Square root.
                        EXPTR(long) p1,         // Base, then remainder.
                        EXPTR(long) p           // Buffer of size *p1/2 + 2.
        )                                       // p1, p may not overlap.
{
  assert (cBigNumberIsFit (p1));                // Check of normalization.
  assert (p1 != p);                             // Check if not overlap.
  size_t n1 = (size_t)(*p1);                    // Number of words.
  assert (p1 [n1] >= 0);                        // Check if non-negative.

  cBigTemp cBigBuf2;                            // Allocate temporary buffer.
  cBigBuf2.checkexpand (n1 + 1);                // +1 for debug check.
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);        // Buffer for subtracter.
  *pp2 = 0;                                     // Initializing.
  cBigNumberFitTo (pp2, n1);                    // Filling by 0.

//      Algorithm.

  long nbit = (cBigNumberBits (p1) + 1) & ~1;
  if ((nbit -= 2) >= 0)
  {

//      Obtaining of the high bit of subtracter.

    {
      size_t nw = (size_t)(nbit / BITS);        // Word.
      long lw = (1L << (nbit % BITS));          // Mask.
      pp2 [nw + 1] += lw;                       // Set bit.
      assert (pp2 [(size_t)(*pp2)] >= 0);       // Check if non-negative.
      p1  [nw + 1] -= lw;                       // Non-normalized subtraction.
      assert (p1  [(size_t)(*p1)]  >= 0);       // Check if non-negative.
    }

//      Obtaining of the other bits of subtracter.
//      Subtrahend and subtracter are not yet normalized,
//      but will be normalized during successive operations.

    while ((nbit -= 2) >= 0)                    // Cycle on lower bits.
    {
      size_t nw = (size_t)(nbit / BITS);        // Word.
      long lw = (1L << (nbit % BITS));          // Mask.
      pp2 [nw + 1] += lw;                       // Set bit.
#ifndef _CBIGNUM_MT
      _cBigNumberSkip = nw + 1;                 // Skip words.
#endif//_CBIGNUM_MT
      if (cBigNumberComp (p1, pp2) < 0)         // Comparison.
      {
        pp2 [nw + 1] -= lw;                     // Unset but.
        cBigNumberMDiv2D (pp2);                 // Bit shift of subtracter.
      }
      else
      {
        cBigNumberMSubD (p1, pp2, 0);           // Correction of remainder.
        assert (p1 [(size_t)(*p1)] >= 0);       // Check if non-negative.
        assert (cBigNumberIsFit (p1));          // Check of normalization.
        pp2 [nw + 1] -= lw;                     // Unset bit.
        cBigNumberMDiv2D (pp2);                 // Bit shift of subtracter.
        pp2 [nw + 1] += lw;                     // Set bit.
      }
      assert (pp2 [(size_t)(*pp2)] >= 0);       // Check if non-negative.
      assert (cBigNumberIsFit (pp2));           // Check of normalization.
    }
  }

  cBigNumberFit (pp2);                          // Normalization of root.
  cBigNumberCopy (pp2, p);                      // Copying of root.
#ifndef _CBIGNUM_MT
  _cBigNumberSkip = 1;                          // Do not skip low 0-words.
#endif//_CBIGNUM_MT

//      Check if square of root obtained is the greatest not less the base.

  assert (cBigNumberComp ((cBigNumberMMul2M (pp2), pp2), p1) >= 0);
}

//================================================
//      Random generator.
//================================================

//      Random generator uses external random function returning
//      unsigned long integers with uniform distribution within
//      the range 0..ULONG_MAX.
//
//      Generator returns uniform-distributed normalized random number,
//      containing limited number of meaning bits.

void    cBigNumberRandom (                      // Random value 0.
                unsigned long (*pfnRand)(),     // Random generator.
                unsigned long lBits,            // Number of meaning bits.
                        EXPTR(long) p           // Buffer of size lBits/BITS+1.
        )
{
  long   l1 = lBits % BITS;                     // Number of bits.
  size_t n1 = (size_t)(lBits / BITS + 1);       // Number of words.

  if (l1 != 0) l1 = (*pfnRand)() & ((unsigned long)~0L >> (BITS-(size_t)l1));
  p [0] = (long)n1;                             // Number of words.
  p [n1] = l1;                                  // The high word.
  while (--n1 != 0) p [n1] = (*pfnRand)();      // Other words.

  cBigNumberFit (p);                            // Normalization.
  assert (p [(size_t)(*p)] >= 0);               // Check if non-negative.
}

#ifdef  __cplusplus
}
#endif//__cplusplus

#endif//_CBIGNUMF_INL
